@@ -1,13 +1,13 @@
-/* @(#)buffer.c	1.164 12/01/01 Copyright 1985, 1995, 2001-2012 J. Schilling */
+/* @(#)buffer.c	1.196 19/03/26 Copyright 1985, 1995, 2001-2019 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)buffer.c	1.164 12/01/01 Copyright 1985, 1995, 2001-2012 J. Schilling";
+	"@(#)buffer.c	1.196 19/03/26 Copyright 1985, 1995, 2001-2019 J. Schilling";
 #endif
 /*
  *	Buffer handling routines
  *
- *	Copyright (c) 1985, 1995, 2001-2012 J. Schilling
+ *	Copyright (c) 1985, 1995, 2001-2019 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -16,6 +16,8 @@ static	UConst char sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
@@ -40,7 +42,6 @@ static	UConst char sccsid[] =
 #include <schily/stdio.h>
 #include <schily/stdlib.h>
 #include <schily/unistd.h>
-#include <schily/libport.h>	/* getpagesize() */
 #include <schily/fcntl.h>
 #include <schily/ioctl.h>
 #include <schily/varargs.h>
@@ -51,12 +52,35 @@ static	UConst char sccsid[] =
 #include "fifo.h"
 #include <schily/string.h>
 #include <schily/wait.h>
-#include <schily/schily.h>
 #include <schily/mtio.h>
 #include <schily/librmt.h>
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
+#include <schily/schily.h>
 #include "starsubs.h"
 
 #include <schily/io.h>		/* for setmode() prototype */
+#include <schily/libport.h>	/* getpagesize() */
+
+#include <schily/nlsdefs.h>
+
+/*
+ * Warning: we need the siginfo_t feature that has been introduced in 1989
+ * with SVr4 and in 1995 with SUSv1. You need a platform that is maintained
+ * since 1995.
+ *
+ * defined(HAVE_SIGPROCMASK) && defined(SA_RESTART) identifies a system that
+ * supports sigaction().
+ *
+ * defined(HAVE_SIGINFO_T) && defined(HAVE_WAITID) is needed in order to get
+ * useful values in siginfo_t. Note that Mac OS X before approx. 2018 e.g.
+ * neither fills in sip->si_pid nor sip->si_code, making siginfo_t useless.
+ */
+#if	defined(HAVE_SIGPROCMASK) && defined(SA_RESTART) && \
+	defined(SA_SIGINFO) && \
+	defined(HAVE_SIGINFO_T) && defined(HAVE_WAITID)
+#define	USE_SIGCLD
+#endif
 
 long	bigcnt	= 0;
 int	bigsize	= 0;		/* Tape block size (may shrink < bigbsize) */
@@ -73,8 +97,8 @@ m_stats	*stats	= &bstat;
 int	pid;
 
 #ifdef	timerclear
-LOCAL	struct	timeval	starttime;
-LOCAL	struct	timeval	stoptime;
+LOCAL	struct	timespec	starttime;
+LOCAL	struct	timespec	stoptime;
 #endif
 
 LOCAL	BOOL	isremote = FALSE;
@@ -93,6 +117,8 @@ extern	char	*tarfiles[];
 extern	int	ntarfiles;
 extern	int	tarfindex;
 extern	BOOL	force_noremote;
+extern	char	*rsh;
+extern	char	*rmt;
 LOCAL	int	lastremote = -1;
 extern	BOOL	multivol;
 extern	char	*newvol_script;
@@ -117,6 +143,9 @@ extern	BOOL	lzoflag;
 extern	BOOL	p7zflag;
 extern	BOOL	xzflag;
 extern	BOOL	lzipflag;
+extern	BOOL	zstdflag;
+extern	BOOL	lzmaflag;
+extern	BOOL	freezeflag;
 extern	char	*compress_prg;
 extern	BOOL	multblk;
 extern	BOOL	partial;
@@ -179,6 +208,10 @@ EXPORT	void	exprstats	__PR((int ret));
 EXPORT	void	excomerrno	__PR((int err, char *fmt, ...)) __printflike__(2, 3);
 EXPORT	void	excomerr	__PR((char *fmt, ...)) __printflike__(1, 2);
 EXPORT	void	die		__PR((int err));
+#ifdef	USE_SIGCLD
+LOCAL	void	cldhandler	__PR((int sig, siginfo_t *sip, void *context));
+LOCAL	void	handlecld	__PR((void));
+#endif
 LOCAL	void	compressopen	__PR((void));
 LOCAL	void	compressclose	__PR((void));
 
@@ -207,6 +240,14 @@ openremote()
 #ifdef	USE_REMOTE
 		isremote = TRUE;
 		rmtdebug(debug);
+		if (rsh)
+			rmtrsh(rsh);
+#ifdef	USE_SSH
+		else
+			rmtrsh("ssh");
+#endif
+		if (rmt)
+			rmtrmt(rmt);
 		rmthostname(host, sizeof (host), tarfiles[tarfindex]);
 		if (debug)
 			errmsgno(EX_BAD, "Remote: %s Host: %s file: %s\n",
@@ -314,8 +355,9 @@ opentape()
 		/*
 		 * XXX Should we add an option that allows to specify O_TRUNC?
 		 */
-		while ((tarf = fileopen(tarfiles[tarfindex],
-						cflag?"rwcub":"rub")) ==
+		while ((tarf = lfilemopen(tarfiles[tarfindex],
+						cflag?"rwcub":"rub",
+						S_IRWALL)) ==
 								(FILE *)NULL) {
 			if (!wready || n++ > 12 ||
 			    (geterrno() != EIO && geterrno() != EBUSY)) {
@@ -345,7 +387,9 @@ opentape()
 	 */
 	if (stats->volno == 1 &&
 	    tape_isreg && !cflag && (!Zflag && !zflag && !bzflag && !lzoflag &&
-	    !p7zflag && !xzflag && !compress_prg)) {
+	    !p7zflag && !xzflag && !lzipflag && !zstdflag && !lzmaflag &&
+	    !freezeflag &&
+	    !compress_prg)) {
 		long	htype;
 		TCB	*ptb;
 
@@ -375,25 +419,40 @@ opentape()
 				bzflag = TRUE;
 				break;
 			case C_LZO:
-				if (!silent) errmsgno(EX_BAD,
+				if (!silent && !print_artype) errmsgno(EX_BAD,
 					"WARNING: Archive is 'lzop' compressed, trying to use the -lzo option.\n");
 				lzoflag = TRUE;
 				break;
 			case C_7Z:
-				if (!silent) errmsgno(EX_BAD,
+				if (!silent && !print_artype) errmsgno(EX_BAD,
 					"WARNING: Archive is '7z' compressed, trying to use the -7z option.\n");
 				p7zflag = TRUE;
 				break;
 			case C_XZ:
-				if (!silent) errmsgno(EX_BAD,
+				if (!silent && !print_artype) errmsgno(EX_BAD,
 					"WARNING: Archive is 'xz' compressed, trying to use the -xz option.\n");
 				xzflag = TRUE;
 				break;
 			case C_LZIP:
-				if (!silent) errmsgno(EX_BAD,
+				if (!silent && !print_artype) errmsgno(EX_BAD,
 					"WARNING: Archive is 'lzip' compressed, trying to use the -lzip option.\n");
 				lzipflag = TRUE;
 				break;
+			case C_ZSTD:
+				if (!silent && !print_artype) errmsgno(EX_BAD,
+					"WARNING: Archive is 'zstd' compressed, trying to use the -zstd option.\n");
+				zstdflag = TRUE;
+				break;
+			case C_LZMA:
+				if (!silent && !print_artype) errmsgno(EX_BAD,
+					"WARNING: Archive is 'lzma' compressed, trying to use the -lzma option.\n");
+				lzmaflag = TRUE;
+				break;
+			case C_FREEZE2:
+				if (!silent && !print_artype) errmsgno(EX_BAD,
+					"WARNING: Archive is 'freeze2' compressed, trying to use the -freeze option.\n");
+				freezeflag = TRUE;
+				break;
 			default:
 				if (!silent) errmsgno(EX_BAD,
 					"WARNING: Unknown compression type %d.\n", cmptype);
@@ -403,7 +462,7 @@ opentape()
 		mtseek((off_t)0, SEEK_SET);
 	}
 	if (Zflag || zflag || bzflag || lzoflag ||
-	    p7zflag || xzflag || lzipflag ||
+	    p7zflag || xzflag || lzipflag || zstdflag || lzmaflag || freezeflag ||
 	    compress_prg) {
 		if (isremote)
 			comerrno(EX_BAD, "Cannot compress remote archives (yet).\n");
@@ -417,8 +476,8 @@ opentape()
 	}
 
 #ifdef	timerclear
-	if (showtime && starttime.tv_sec == 0 && starttime.tv_usec == 0 &&
-	    gettimeofday(&starttime, (struct timezone *)0) < 0)
+	if (showtime && starttime.tv_sec == 0 && starttime.tv_nsec == 0 &&
+	    getnstimeofday(&starttime) < 0)
 		comerr("Cannot get starttime\n");
 #endif
 }
@@ -453,7 +512,7 @@ EXPORT void
 changetape(donext)
 	BOOL	donext;
 {
-	char	ans[2];
+	char	ans[3];
 	int	nextindex;
 
 	if (donext) {
@@ -516,10 +575,20 @@ changetape(donext)
 				stats->volno, tarfiles[nextindex]);
 		system(scrbuf);
 	} else {
+		int	len;
+
 		errmsgno(EX_BAD, "Mount volume #%d on '%s' and hit <RETURN>",
 			stats->volno, tarfiles[nextindex]);
-		fgetline(tty, ans, sizeof (ans));
-		if (feof(tty))
+		ans[0] = '\n';
+		len = fgetstr(tty, ans, sizeof (ans));
+		if (len > 0 && ans[len-1] != '\n') {
+			while (getc(tty) != '\n') {
+				if (feof(tty) || ferror(tty))
+					break;
+			}
+		}
+
+		if (ttyerr(tty))
 			exit(1);
 	}
 	tarfindex = nextindex;
@@ -602,15 +671,17 @@ extern	m_head	*mp;
 		"Panic: trying to write more than bs (%d > %d)!\n",
 		amount, bigsize);
 	}
-	mp->chreel = TRUE;
 #ifdef	FIFO
 	if (use_fifo) {
+		mp->chreel = TRUE;
+
 		/*
 		 * Make sure the put side of the FIFO is waiting either on
 		 * mp->iblocked (because the FIFO is full) or on mp->reelwait
 		 * before temporary disabling the FIFO during media change.
 		 */
-		while (mp->iblocked == FALSE && mp->reelwait == FALSE) {
+		while ((mp->eflags & FIFO_EXIT) == 0 &&
+		    mp->iblocked == FALSE && mp->reelwait == FALSE) {
 			usleep(100000);
 		}
 	}
@@ -650,9 +721,9 @@ extern	m_head	*mp;
 	bigcnt = ocnt;
 	use_fifo = ofifo;
 	active = FALSE;
-	mp->chreel = FALSE;
 #ifdef	FIFO
 	if (use_fifo) {
+		mp->chreel = FALSE;
 		fifo_reelwake();
 	}
 #endif
@@ -821,8 +892,11 @@ initbuf(nblocks)
 	 * the shared memory in the FIFO while trying to detect the archive
 	 * format and byte swapping in read/extract modes.
 	 * Note that -r and -u currently disable the FIFO.
+	 * In case that we enable the FIFO for -r and -u, we need to add
+	 * another exception here in order to have space to remember the last
+	 * block of significant data that needs to be modified to append.
 	 */
-	if (!use_fifo || cvolhdr) {
+	if (!use_fifo || cvolhdr || rflag || uflag) {
 		int	pagesize = getpagesize();
 
 		/*
@@ -834,8 +908,12 @@ initbuf(nblocks)
 		 * "bufsize" may be modified by initfifo() in the FIFO case,
 		 * so we use "bigsize" for the extra multivol buffer to
 		 * avoid allocating an unneeded huge amount of data here.
+		 * In "replace" or "update" mode, we also may need to
+		 * save/restore * the buffer for the tape record when doing
+		 * EOF detection. As this space is needed at a different
+		 * time, it may be shared with the extra multivol buffer.
 		 */
-		if (cvolhdr)
+		if (cvolhdr || rflag || uflag)
 			bigsize *= 2;
 
 		/*
@@ -850,7 +928,7 @@ initbuf(nblocks)
 		fillbytes(bigbuf, bigsize, '\0');
 		fillbytes(&bigbuf[bigsize], 10, 'U');
 
-		if (cvolhdr) {
+		if (cvolhdr || rflag || uflag) {
 			bigsize /= 2;
 			bigbase = bigbuf;
 			bigbuf = bigptr = &bigbase[bigsize];
@@ -881,7 +959,8 @@ markeof()
 
 	if (debug) {
 		error("Blocks: %lld\n", tblocks());
-		error("bigptr - bigbuff: %lld %p %p %p lastsize: %ld\n",
+		error(
+		"bigptr - bigbuff: %lld bigbuf: %p bigptr: %p eofptr: %p lastsize: %ld\n",
 			(Llong)(bigptr - bigbuf),
 			(void *)bigbuf, (void *)bigptr, (void *)eofptr,
 			stats->lastsize);
@@ -905,10 +984,10 @@ marktcb(addr)
 #ifdef	FIFO
 	bit = addr - mp->base;
 	if (bit % TBLOCK)		/* Remove this paranoia test in future. */
-		error("TCB offset not mudulo 512.\n");
+		errmsgno(EX_BAD, "TCB offset not mudulo 512.\n");
 	bit /= TBLOCK;
 	if (bit_test(mp->bmap, bit))	/* Remove this paranoia test in future. */
-		error("Bit %d is already set.\n", bit);
+		errmsgno(EX_BAD, "Bit %d is already set.\n", bit);
 	bit_set(mp->bmap, bit);
 #endif
 }
@@ -1388,7 +1467,8 @@ buf_drain()
 {
 #ifdef	FIFO
 	if (use_fifo) {
-		fifo_oflush();
+		fifo_oflush();	/* Set FIFO_MEOF flag and wake other side */
+		fifo_oclose();	/* Close sync pipe to finally wake other side */
 		wait(0);
 	}
 #endif
@@ -1511,7 +1591,7 @@ extern	long	hdrtype;
 /*
  * Backspace tape or medium to prepare it for appending to an archive.
  * Note that this currently only handles TAR archives and that even then it
- * will not work if the TAPE recordr size is < 2*TBLOCK (1024 bytes).
+ * will not work if the TAPE record size is < 2*TBLOCK (1024 bytes).
  */
 EXPORT void
 backtape()
@@ -1639,7 +1719,7 @@ prstats()
 	int	per;
 #ifdef	timerclear
 	int	sec;
-	int	usec;
+	int	nsec;
 	int	tmsec;
 #endif
 	char	*p;
@@ -1661,7 +1741,7 @@ prstats()
 		return;
 
 #ifdef	timerclear
-	if (showtime && gettimeofday(&stoptime, (struct timezone *)0) < 0)
+	if (showtime && getnstimeofday(&stoptime) < 0)
 		comerr("Cannot get stoptime\n");
 #endif
 #ifdef	FIFO
@@ -1702,11 +1782,11 @@ prstats()
 		Llong	kbs;
 
 		sec = stoptime.tv_sec - starttime.tv_sec;
-		usec = stoptime.tv_usec - starttime.tv_usec;
-		tmsec = sec*1000 + usec/1000;
-		if (usec < 0) {
+		nsec = stoptime.tv_nsec - starttime.tv_nsec;
+		tmsec = sec*1000 + nsec/1000000;
+		if (nsec < 0) {
 			sec--;
-			usec += 1000000;
+			nsec += 1000000000;
 		}
 		if (tmsec == 0)
 			tmsec++;
@@ -1714,9 +1794,12 @@ prstats()
 		kbs = kbytes*(Llong)1000/tmsec;
 
 		errmsgno(EX_BAD, "Total time %d.%03dsec (%lld kBytes/sec)\n",
-				sec, usec/1000, kbs);
+				sec, nsec/1000000, kbs);
 	}
 #endif
+#ifdef	DBG_MALLOC
+	aprintlist(stdout, 1);
+#endif
 }
 
 EXPORT BOOL
@@ -1734,6 +1817,9 @@ checkerrs()
 	    xstats.s_isspecial	||
 	    xstats.s_sizeerrs	||
 	    xstats.s_chdir	||
+	    xstats.s_iconv	||
+	    xstats.s_id		||
+	    xstats.s_time	||
 
 	    xstats.s_settime	||
 	    xstats.s_security	||
@@ -1748,16 +1834,25 @@ checkerrs()
 	    xstats.s_setxattr	||
 #endif
 	    xstats.s_setmodes	||
-	    xstats.s_restore) {
+	    xstats.s_restore	||
+	    xstats.s_compress	||
+	    xstats.s_hardeof	||
+	    xstats.s_substerrs	||
+	    xstats.s_selinuxerrs) {
 		if (nowarn || no_stats || (pid == 0) /* child */)
 			return (TRUE);
 
 		errmsgno(EX_BAD, "The following problems occurred during archive processing:\n");
-		errmsgno(EX_BAD, "Cannot: stat %d, open %d, read/write %d, chdir %d. Size changed %d.\n",
+		errmsgno(EX_BAD, "Cannot: stat %d, open %d, read/write %d, chdir %d, iconv %d.\n",
 				xstats.s_staterrs,
 				xstats.s_openerrs,
 				xstats.s_rwerrs,
 				xstats.s_chdir,
+				xstats.s_iconv);
+		if (xstats.s_id || xstats.s_time)
+			errmsgno(EX_BAD, "Range errors: uid/gid %d, time: %d.\n",
+				xstats.s_id, xstats.s_time);
+		errmsgno(EX_BAD, "Size changed %d.\n",
 				xstats.s_sizeerrs);
 		errmsgno(EX_BAD, "Missing links %d, Name too long %d, File too big %d, Not dumped %d.\n",
 				xstats.s_misslinks,
@@ -1786,9 +1881,27 @@ checkerrs()
 			errmsgno(EX_BAD, "Cannot get xattr: %d set xattr: %d.\n",
 				xstats.s_getxattr,
 				xstats.s_setxattr);
+#endif
+#ifdef USE_SELINUX
+		if (xstats.s_selinuxerrs)
+			errmsgno(EX_BAD, "Cannot set SELinux security context: %d.\n",
+				xstats.s_selinuxerrs);
 #endif
 		if (xstats.s_restore)
 			errmsgno(EX_BAD, "Problems with restore database.\n");
+		if (xstats.s_compress)
+			errmsgno(EX_BAD, "Problems with compress program.\n");
+		if (xstats.s_substerrs)
+			errmsgno(EX_BAD,
+				"%d Problem(s) with path substitution.\n",
+				xstats.s_substerrs);
+		if (xstats.s_hardeof)
+			errmsgno(EX_BAD, "Hard EOF on input.\n");
+
+		if (xstats.s_security)
+			errmsgno(EX_BAD, "See option -.. on why some files have been skipped.\n");
+		if (xstats.s_lsecurity)
+			errmsgno(EX_BAD, "See option -secure-links on why some links have been skipped.\n");
 		return (TRUE);
 	}
 	return (FALSE);
@@ -1800,6 +1913,8 @@ exprstats(ret)
 {
 	prstats();
 	checkerrs();
+	if (use_fifo)
+		fifo_exit(ret);
 	exit(ret);
 }
 
@@ -1822,7 +1937,7 @@ excomerrno(err, fmt, va_alist)
 #else
 	va_start(args);
 #endif
-	errmsgno(err, "%r", fmt, args);
+	errmsgno(err, "%r", _(fmt), args);
 	va_end(args);
 #ifdef	FIFO
 	fifo_exit(err);
@@ -1850,7 +1965,7 @@ excomerr(fmt, va_alist)
 #else
 	va_start(args);
 #endif
-	errmsgno(err, "%r", fmt, args);
+	errmsgno(err, "%r", _(fmt), args);
 	va_end(args);
 #ifdef	FIFO
 	fifo_exit(err);
@@ -1873,6 +1988,45 @@ die(err)
 #if	defined(SIGDEFER) || defined(SVR4)
 #define	signal	sigset
 #endif
+
+LOCAL	pid_t	compresspid;
+
+#ifdef	USE_SIGCLD
+LOCAL void
+cldhandler(sig, sip, context)
+	int		sig;
+	siginfo_t	*sip;
+	void		*context;
+{
+	if (sip->si_pid != compresspid)
+		return;
+
+	if (sip->si_status != 0 || sip->si_code != CLD_EXITED)
+		xstats.s_compress++;
+
+	if (sip->si_status != 0 && sip->si_code == CLD_EXITED)
+		errmsgno(EX_BAD,
+		"Compress program exited with status %d.\n",
+			sip->si_status);
+	else if (sip->si_status != 0)
+		errmsgno(EX_BAD,
+		"Compress program died with signal %d.\n",
+			sip->si_status);
+}
+
+LOCAL void
+handlecld()
+{
+	struct sigaction sa;
+
+	sa.sa_sigaction = cldhandler;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART|SA_SIGINFO;
+
+	sigaction(SIGCHLD, &sa, NULL);
+}
+#endif	/* USE_SIGCLD */
+
 LOCAL void
 compressopen()
 {
@@ -1895,6 +2049,12 @@ compressopen()
 		zip_prog = "xz";
 	else if (lzipflag)
 		zip_prog = "lzip";
+	else if (zstdflag)
+		zip_prog = "zstd";
+	else if (lzmaflag)
+		zip_prog = "lzma";
+	else if (freezeflag)
+		zip_prog = "freeze";
 
 	multblk = TRUE;
 
@@ -1944,6 +2104,9 @@ compressopen()
 #else
 	if (fpipe(pp) == 0)
 		comerr("Compress pipe failed\n");
+#ifdef	USE_SIGCLD
+	handlecld();
+#endif
 	mypid = fork();
 	if (mypid < 0)
 		comerr("Compress fork failed\n");
@@ -1965,14 +2128,21 @@ compressopen()
 #endif
 
 		/* We don't want to see errors */
-		null = fileopen("/dev/null", "rw");
+		null = lfilemopen("/dev/null", "rw", S_IRWALL);
+		if (null == NULL) {
+			errmsg("Cannot open '%s'.\n", "/dev/null");
+			goto err;
+		}
 
 		if (cflag)
 			fexecl(zip_prog, pp[0], tarf, null, zip_prog, flg, (char *)NULL);
 		else
 			fexecl(zip_prog, tarf, pp[1], null, zip_prog, "-d", (char *)NULL);
+err:
 		errmsg("Compress: exec of '%s' failed\n", zip_prog);
 		_exit(-1);
+	} else {
+		compresspid = mypid;
 	}
 	fclose(tarf);
 	if (cflag) {
@@ -1993,23 +2163,6 @@ LOCAL void
 compressclose()
 {
 #ifdef HAVE_FORK
-	char	*zip_prog = "gzip";
-
-	if (compress_prg)
-		zip_prog = compress_prg;
-	else if (bzflag)
-		zip_prog = "bzip2";
-	else if (Zflag)
-		zip_prog = "compress";
-	else if (lzoflag)
-		zip_prog = "lzop";
-	else if (p7zflag)
-		zip_prog = "p7zip";
-	else if (xzflag)
-		zip_prog = "xz";
-	else if (lzipflag)
-		zip_prog = "lzip";
-
 #ifdef __DJGPP__
 	if (cflag) {
 		if (compress_tmpf) {
@@ -2017,6 +2170,28 @@ compressclose()
 			FILE	*zipf;
 			int	cnt = -1;
 			char	buf[8192];
+			char	*zip_prog = "gzip";
+
+			if (compress_prg)
+				zip_prog = compress_prg;
+			else if (bzflag)
+				zip_prog = "bzip2";
+			else if (Zflag)
+				zip_prog = "compress";
+			else if (lzoflag)
+				zip_prog = "lzop";
+			else if (p7zflag)
+				zip_prog = "p7zip";
+			else if (xzflag)
+				zip_prog = "xz";
+			else if (lzipflag)
+				zip_prog = "lzip";
+			else if (zstdflag)
+				zip_prog = "zstd";
+			else if (lzmaflag)
+				zip_prog = "lzma";
+			else if (freezeflag)
+				zip_prog = "freeze";
 
 			js_snprintf(zip_cmd, sizeof (zip_cmd), "%s.exe", zip_prog);
 
