@@ -1,11 +1,11 @@
-/* @(#)create.c	1.135 14/02/05 Copyright 1985, 1995, 2001-2014 J. Schilling */
+/* @(#)create.c	1.156 19/03/11 Copyright 1985, 1995, 2001-2019 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)create.c	1.135 14/02/05 Copyright 1985, 1995, 2001-2014 J. Schilling";
+	"@(#)create.c	1.156 19/03/11 Copyright 1985, 1995, 2001-2019 J. Schilling";
 #endif
 /*
- *	Copyright (c) 1985, 1995, 2001-2014 J. Schilling
+ *	Copyright (c) 1985, 1995, 2001-2019 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -31,10 +31,14 @@ static	UConst char sccsid[] =
 #include <schily/unistd.h>
 #include <schily/dirent.h>
 #include <schily/string.h>
+#include <schily/jmpdefs.h>	/* For __fjmalloc() */
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
 #include <schily/schily.h>
 #include <schily/idcache.h>
 #include "restore.h"
 #ifdef	USE_FIND
+#include <schily/fcntl.h>	/* For AT_FDCWD */
 #include <schily/stat.h>	/* Fuer stat_to_info() in starsubs.h */
 #include <schily/walk.h>
 #include <schily/find.h>
@@ -48,8 +52,8 @@ static	UConst char sccsid[] =
 
 #ifdef	OWN_ACLTEXT
 #if	defined(UNIXWARE) && defined(HAVE_ACL)
-#	define	HAVE_SUN_ACL
-#	define	HAVE_ANY_ACL
+#define	HAVE_SUN_ACL
+#define	HAVE_ANY_ACL
 #endif
 #endif
 /*
@@ -60,9 +64,9 @@ static	UConst char sccsid[] =
  * star.h at all.
  * HAVE_HP_ACL is currently not included in HAVE_ANY_ACL.
  */
-#	ifndef	HAVE_ANY_ACL
-#	undef	USE_ACL		/* Do not try to get or set ACLs */
-#	endif
+#ifndef	HAVE_ANY_ACL
+#undef	USE_ACL		/* Do not try to get or set ACLs */
+#endif
 #endif
 
 struct pdirs {
@@ -164,10 +168,11 @@ LOCAL	mode_t	statmode = _BAD_MODE;
 
 EXPORT	void	checklinks	__PR((void));
 LOCAL	int	take_file	__PR((char *name, FINFO *info));
-EXPORT	int	_fileopen	__PR((char *name, char *mode));
 EXPORT	int	_fileread	__PR((int *fp, void *buf, int len));
 EXPORT	void	create		__PR((char *name, BOOL Hflag, BOOL forceadd));
-LOCAL	void	createi		__PR((char *sname, char *name, int namlen, FINFO *info, struct pdirs *last));
+LOCAL	void	createi		__PR((char *sname, char *name, int namlen,
+					FINFO *info, pathstore_t *pathp,
+					struct pdirs *last));
 EXPORT	void	createlist	__PR((void));
 LOCAL	BOOL	get_metainfo	__PR((char *line));
 EXPORT	BOOL	read_symlink	__PR((char *sname, char *name, FINFO *info, TCB *ptb));
@@ -184,10 +189,12 @@ EXPORT	void	put_file	__PR((int *fp, FINFO *info));
 EXPORT	void	cr_file		__PR((FINFO *info,
 					int (*)(void *, char *, int),
 					void *arg, int amt, char *text));
-LOCAL	void	put_dir		__PR((char *sname, char *dname, int namlen, FINFO *info,
-								TCB *ptb, struct pdirs *last));
+LOCAL	void	put_dir		__PR((char *sname, char *dname, int namlen,
+					FINFO *info, TCB *ptb,
+					pathstore_t *pathp, struct pdirs *last));
+LOCAL	void	toolong		__PR((char *dname, char	*name, size_t len));
 LOCAL	BOOL	checkdirexclude	__PR((char *sname, char *name, int namlen, FINFO *info));
-EXPORT	BOOL	checkexclude	__PR((char *sname, char *name, int namlen, FINFO *info));
+LOCAL	BOOL	checkexclude	__PR((char *sname, char *name, int namlen, FINFO *info));
 
 #ifdef	USE_FIND
 EXPORT	int	walkfunc	__PR((char *nm, struct stat *fs, int type, struct WALK *state));
@@ -286,8 +293,8 @@ take_file(name, info)
 	} else if (dumplevel > 0) {
 		/*
 		 * For now, we cannot reliably deal with sub-second granularity
-		 * on all platforms. For this reason, some files to be on two
-		 * incrementals to make sure not to miss them completely.
+		 * on all platforms. For this reason, we let some files be on
+		 * two incrementals to make sure not to miss them completely.
 		 */
 		if (info->f_mtime >= Newer.tv_sec) {
 			/* EMPTY */
@@ -302,9 +309,22 @@ take_file(name, info)
 	} else if (Newer.tv_sec && (Ctime ? info->f_ctime:info->f_mtime) <=
 								Newer.tv_sec) {
 		/*
-		 * XXX nsec beachten wenn im Archiv!
+		 * First the easy case: Don't take file if seconds are
+		 * less than in the reference file.
 		 */
-		return (FALSE);
+		if ((Ctime ? info->f_ctime:info->f_mtime) < Newer.tv_sec)
+			return (FALSE);
+		/*
+		 * If we do not have nanoseconds, we are done.
+		 */
+		if ((info->f_flags & F_NSECS) == 0)
+			return (FALSE);
+		/*
+		 * Here we know that we have nanoseconds and that seconds
+		 * are equal, so finally check nanoseconds.
+		 */
+		if ((Ctime ? info->f_cnsec:info->f_mnsec) <= Newer.tv_nsec)
+			return (FALSE);
 	} else if (uflag && !update_newer(info)) {
 		return (FALSE);
 	} else if (!multivol &&
@@ -382,28 +402,6 @@ take_file(name, info)
 	return (TRUE);
 }
 
-int
-_fileopen(name, smode)
-	char	*name;
-	char	*smode;
-{
-	int	ret;
-	int	omode = 0;
-	int	flag = 0;
-
-	if (!_cvmod(smode, &omode, &flag))
-		return (-1);
-
-retry:
-	if ((ret = _openfd(name, omode)) < 0) {
-		if (geterrno() == EINTR)
-			goto retry;
-		return (-1);
-	}
-
-	return (ret);
-}
-
 int
 _fileread(fp, buf, len)
 	register int	*fp;
@@ -476,18 +474,20 @@ create(name, Hflag, forceadd)
 	if (forceadd)
 		info->f_flags |= F_FORCE_ADD;
 	paxfollow = opaxfollow;
-	createi(name, name, strlen(name), info, (struct pdirs *)0);
+	createi(name, name, strlen(name), info,
+		(pathstore_t *)0, (struct pdirs *)0);
 }
 
 LOCAL void
-createi(sname, name, namlen, info, last)
+createi(sname, name, namlen, info, pathp, last)
 		char	*sname;
 	register char	*name;
 		int	namlen;
 	register FINFO	*info;
+		pathstore_t *pathp;
 		struct pdirs *last;
 {
-		char	lname[PATH_MAX+1];
+		char	lname[PATH_MAX+1];  /* This limit cannot be overruled */
 		TCB	tb;
 	register TCB	*ptb		= &tb;
 		int	fd		= -1;
@@ -505,7 +505,7 @@ createi(sname, name, namlen, info, last)
 
 #ifdef	nonono_NICHT_BEI_CREATE	/* XXX */
 	if (!abs_path &&	/* XXX VVV siehe skip_slash() */
-		(info->f_name[0] == '/' /* || info->f_lname[0] == '/' */))
+		(info->f_name[0] == '/' /* || info->f_lname[0] == '/' */)) {
 		skip_slash(info);
 		info->f_namelen -= info->f_name - name;
 		if (info->f_namelen == 0) {
@@ -524,7 +524,7 @@ createi(sname, name, namlen, info, last)
 	if (do_subst && subst(info)) {
 		if (info->f_name[0] == '\0') {
 			if (verbose)
-			fprintf(vpr,
+			fgtprintf(vpr,
 				"'%s' substitutes to null string, skipping ...\n",
 							name);
 			return;
@@ -540,7 +540,8 @@ createi(sname, name, namlen, info, last)
 			info->f_gid = statgid;
 	}
 
-	if (!(dirmode && is_dir(info)) &&
+	if (verbose <= 1 &&
+	    !(dirmode && is_dir(info)) &&
 				(info->f_namelen <= props.pr_maxsname)) {
 		/*
 		 * Allocate TCB from the buffer to avoid copying TCB
@@ -551,6 +552,10 @@ createi(sname, name, namlen, info, last)
 		 * With very long names we will have to write out
 		 * other data before we can write the TCB, so we cannot
 		 * alloc tcb from buffer too.
+		 * If we are creating a long listing while archiving, the
+		 * TCB will be overwritten by info_to_xhdr() and this would
+		 * overwrite username/groupname that is later needed for
+		 * vprint(), so we cannot allocate TCB from the buffer here.
 		 */
 		if ((ptb = (TCB *)get_block(props.pr_hdrsize)) == NULL)
 			ptb = &tb;
@@ -584,12 +589,12 @@ createi(sname, name, namlen, info, last)
 			put_tcb(ptb, info);
 			vprint(info);
 		} else {
-			put_dir(sname, name, namlen, info, ptb, last);
+			put_dir(sname, name, namlen, info, ptb, pathp, last);
 		}
 	} else if (take_file(name, info) <= 0) {	/* < TRUE */
 		return;
 	} else if (interactive && !ia_change(ptb, info)) {
-		fprintf(vpr, "Skipping ...\n");
+		fgtprintf(vpr, "Skipping ...\n");
 #ifdef	__PRE_CPIO__
 	} else if (is_symlink(info) && !read_symlink(sname, name, info, ptb)) {
 		/* EMPTY */
@@ -632,7 +637,7 @@ createi(sname, name, namlen, info, last)
 		return;
 
 	} else if (is_file(info) && info->f_size != 0 && !nullout &&
-				(fd = _fileopen(sname, "rb")) < 0) {
+				(fd = _lfileopen(sname, "rb")) < 0) {
 		if (!errhidden(E_OPEN, name)) {
 			if (!errwarnonly(E_OPEN, name))
 				xstats.s_openerrs++;
@@ -693,7 +698,7 @@ createi(sname, name, namlen, info, last)
 		 */
 		do_sparse = sparse && (props.pr_flags & PR_SPARSE);
 		if (do_sparse && nullout &&
-				(fd = _fileopen(sname, "rb")) < 0) {
+				(fd = _lfileopen(sname, "rb")) < 0) {
 			if (!errhidden(E_OPEN, name)) {
 				if (!errwarnonly(E_OPEN, name))
 					xstats.s_openerrs++;
@@ -742,7 +747,7 @@ createlist()
 {
 	register int	nlen;
 		char	*name;
-		int	nsize = PATH_MAX+1+512;	/* wegen laenge !!! */
+		size_t	nsize = PATH_MAX;
 
 	/*
 	 * We need at least PATH_MAX+1 and add 512 to get better messages below
@@ -750,27 +755,23 @@ createlist()
 	name = ___malloc(nsize, "name buffer");
 
 	for (nlen = 1; nlen >= 0; ) {
-		if ((nlen = readnull ? ngetline(listf, name, nsize) :
-					fgetline(listf, name, nsize)) < 0)
+		if ((nlen = getdelim(&name, &nsize,
+				readnull ? '\0' : '\n', listf)) < 0)
 			break;
+
 		if (nlen == 0)
 			continue;
+		if (!readnull) {
+			if (name[nlen-1] == '\n')
+				name[--nlen] = '\0';
+			if (nlen == 0)
+				continue;
+		}
 		if (pkglist) {
 			if (!get_metainfo(name))
 				continue;
 			nlen = strlen(name);
 		}
-		if (nlen >= PATH_MAX) {
-			if (!errhidden(E_NAMETOOLONG, name)) {
-				if (!errwarnonly(E_NAMETOOLONG, name))
-					xstats.s_toolong++;
-				errmsgno(EX_BAD,
-				"%s: Name too long (%d >= %d).\n",
-							name, nlen, PATH_MAX);
-				(void) errabort(E_NAMETOOLONG, name, TRUE);
-			}
-			continue;
-		}
 		if (intr)
 			break;
 		curfs = NODEV;
@@ -841,7 +842,7 @@ read_symlink(sname, name, info, ptb)
 	info->f_lname[0] = '\0';
 
 #ifdef	HAVE_READLINK
-	if ((len = readlink(sname, info->f_lname, PATH_MAX)) < 0) {
+	if ((len = lreadlink(sname, info->f_lname, PATH_MAX)) < 0) {
 		if (!errhidden(E_READLINK, name)) {
 			if (!errwarnonly(E_READLINK, name))
 				xstats.s_rwerrs++;
@@ -979,6 +980,9 @@ xcpio_link(info)
 
 /*
  * Archive all cpio CRC links.
+ *
+ * We are called if our caller believes that the last link for a specific file
+ * has been encountered.
  */
 LOCAL BOOL
 acpio_link(info)
@@ -990,13 +994,15 @@ acpio_link(info)
 		int	namelen = info->f_namelen;
 		char	*name = info->f_name;
 		char	*lname = info->f_lname;
-		off_t	size = info->f_size;
-		off_t	rsize = info->f_rsize;
+		off_t	size = info->f_size;	/* real size of file	    */
+#ifdef	__what_people_would_expect__
+		off_t	rsize = info->f_rsize;	/* size as archived on tape */
+#endif
 
 	if ((lp = find_link(info)) != NULL) {
 		/*
 		 * We come here if the link count increases and we need to
-		 * archive more "files" as expected.
+		 * archive more "files" than expected.
 		 */
 		if ((lp->l_flags & L_DATA) != 0)
 			return (FALSE);
@@ -1036,8 +1042,11 @@ acpio_link(info)
 	info->f_name  = name;
 	info->f_lname = lname;
 	info->f_size  = size;
+#ifdef	__what_people_would_expect__
 	info->f_rsize = rsize;
-	info->f_rsize = size;
+#else
+	info->f_rsize = size;	/* Achieve that the last link archives data */
+#endif
 	info->f_xftype = info->f_rxftype;
 
 	info->f_flags &= ~F_TCB_BUF;
@@ -1091,6 +1100,7 @@ flush_link(lp)
 		BOOL	did_stat;
 		char	*name;
 
+	finfo.f_flags = 0;
 	finfo.f_flags &= ~F_TCB_BUF;
 	if ((ptb = (TCB *)get_block(props.pr_hdrsize)) == NULL)
 		ptb = &tb;
@@ -1125,7 +1135,7 @@ flush_link(lp)
 	finfo.f_tcb = ptb;
 	fd = -1;
 	if (is_file(&finfo) && finfo.f_size != 0 && !nullout &&
-				(fd = _fileopen(name, "rb")) < 0) {
+				(fd = _lfileopen(name, "rb")) < 0) {
 		if (!errhidden(E_OPEN, name)) {
 			if (!errwarnonly(E_OPEN, name))
 				xstats.s_openerrs++;
@@ -1246,7 +1256,7 @@ read_link(name, namlen, info, ptb)
 				"Cannot alloc new link name for '%s'.\n",
 					name);
 			} else {
-				strcpy(ln->l_lname, name);
+				strlcpy(ln->l_lname, name, namlen+1);
 				ln->l_lnext = lp->l_lnames;
 				lp->l_lnames = ln;
 			}
@@ -1270,7 +1280,7 @@ read_link(name, namlen, info, ptb)
 			lp->l_flags = L_ISDIR;
 		else
 			lp->l_flags = 0;
-		strcpy(lp->l_name, name);
+		strlcpy(lp->l_name, name, namlen+1);
 
 		if ((info->f_flags & F_EXTRACT) == 0) {
 			/*
@@ -1424,12 +1434,13 @@ extern	m_stats	*stats;
 #define	newfs(i)	((i)->f_dev != curfs)
 
 LOCAL void
-put_dir(sname, dname, namlen, info, ptb, last)
+put_dir(sname, dname, namlen, info, ptb, pathp, last)
 		char	*sname;
 	register char	*dname;
 	register int	namlen;
 	FINFO	*info;
 	TCB	*ptb;
+	pathstore_t	*pathp;
 	struct pdirs	*last;
 {
 	static	int	depth	= -10;
@@ -1441,9 +1452,8 @@ put_dir(sname, dname, namlen, info, ptb, last)
 #ifdef	HAVE_SEEKDIR
 		long	offset	= 0L;
 #endif
-		char	fname[PATH_MAX+1];	/* XXX */
+	pathstore_t	path;
 	register char	*name;
-	register char	*xdname;
 		int	xlen;
 		BOOL	putdir = FALSE;
 		BOOL	direrr = FALSE;
@@ -1451,6 +1461,7 @@ put_dir(sname, dname, namlen, info, ptb, last)
 		char	*dp = NULL;
 	struct pdirs	thisd;
 	struct pdirs	*pd = last;
+		BOOL	didslash = FALSE;
 
 	if (!dinit) {
 #ifdef	_SC_OPEN_MAX
@@ -1498,7 +1509,22 @@ put_dir(sname, dname, namlen, info, ptb, last)
 			ino_t	*ino = NULL;
 			int	nents;
 
-			dp = fetchdir(sname, &nents, &dlen, dodump?&ino:NULL);
+			d = lopendir(sname);
+			if (d) {
+				dp = dfetchdir(d, sname,
+					    &nents, &dlen, dodump?&ino:NULL);
+#if	defined(HAVE_FSTATAT) && defined(HAVE_DIRFD)
+				if (depth <= 0) {
+					closedir(d);
+					d = NULL;
+				}
+#else
+				closedir(d);
+				d = NULL;
+#endif
+			} else {
+				dp = NULL;
+			}
 			if (dp == NULL) {
 				if (!errhidden(E_OPEN, dname)) {
 					if (!errwarnonly(E_OPEN, dname))
@@ -1517,7 +1543,7 @@ put_dir(sname, dname, namlen, info, ptb, last)
 				 */
 				dlen--;
 			}
-		} else if (!(d = opendir(sname))) {
+		} else if (!(d = lopendir(sname))) {
 			if (!errhidden(E_OPEN, dname)) {
 				if (!errwarnonly(E_OPEN, dname))
 					xstats.s_openerrs++;
@@ -1530,7 +1556,7 @@ put_dir(sname, dname, namlen, info, ptb, last)
 	depth--;
 	if (!nodir) {
 		if (interactive && !ia_change(ptb, info)) {
-			fprintf(vpr, "Skipping ...\n");
+			fgtprintf(vpr, "Skipping ...\n");
 			if (d)
 				closedir(d);
 			if (dp)
@@ -1567,67 +1593,84 @@ put_dir(sname, dname, namlen, info, ptb, last)
 	}
 
 	if (!nodesc && (!nomount || !newfs(info))) {
+		if (sname != dname)
+			comerrno(EX_BAD, "Panic: put_dir(): sname != dname\n");
 
-		strcpy(fname, dname);
-		xdname = &fname[namlen];
-		if (namlen && xdname[-1] != '/') {
+		name = dname;
+		if (pathp == NULL) {
+			if (name[0] == '.' && name[1] == '/') {
+				for (name++; name[0] == '/'; name++)
+					/* LINTED */
+					;
+				namlen -= name - dname;
+			}
+			if (name[0] == '.' && name[1] == '\0') {
+				name++;
+				namlen--;
+			}
+			pathp = &path;
+			if (init_pspace(PS_STDERR, pathp) < 0 ||
+			    strlcpy_pspace(PS_STDERR, pathp, name, namlen) < 0) {
+				toolong("", name, namlen);
+				goto out;
+			}
+		}
+		if (namlen && pathp->ps_path[pathp->ps_tail - 1] != '/') {
+			char	*p;
+
+			if ((pathp->ps_tail+2) > pathp->ps_size) {
+				if (incr_pspace(PS_STDERR, pathp, 2) < 0) {
+					toolong("", name, namlen);
+					goto out;
+				}
+			}
+			p = &pathp->ps_path[pathp->ps_tail++];
+			*p++ = '/';
+			*p++ = '\0';
 			namlen++;
-			*xdname++ = '/';
+			didslash = TRUE;
 		}
 
 		while (!intr) {
-			if (d) {
+			char	*snm;
+
+			if (lowmem) {
 				if ((dir = readdir(d)) == NULL)
 					break;
 				if (streql(dir->d_name, ".") ||
 						streql(dir->d_name, ".."))
 					continue;
-				name = dir->d_name;
-				xlen = namlen + strlen(name);
+				snm = dir->d_name;
+				xlen = strlen(snm);
 			} else {
 				if (dlen <= 0)
 					break;
 
-				name = &dp[1];
-				xlen = strlen(name);
+				snm = &dp[1];
+				xlen = strlen(snm);
 				dp += xlen + 2;
 				dlen -= xlen + 2;
-				xlen += namlen;
 			}
 
-			if (xlen > PATH_MAX) {
-				char	xname[2*PATH_MAX+1];
-
-				*xdname = '\0';
-				js_snprintf(xname, sizeof (xname), "%s%s",
-					fname, name);
-				if (!errhidden(E_NAMETOOLONG, xname)) {
-					if (!errwarnonly(E_NAMETOOLONG, xname))
-						xstats.s_toolong++;
-					errmsgno(EX_BAD,
-					"%s%s: Name too long (%d > %d).\n",
-						fname, name,
-						xlen, PATH_MAX);
-					(void) errabort(E_NAMETOOLONG, xname,
-									TRUE);
+			if ((namlen+xlen+1) >= pathp->ps_size) {
+				if (grow_pspace(PS_STDERR, pathp, namlen+xlen+1) < 0) {
+					toolong(pathp->ps_path, snm, namlen+xlen);
+					goto out;
 				}
-				continue;
 			}
+			strcpy(&pathp->ps_path[pathp->ps_tail], snm);
+			name = pathp->ps_path;
 
-			strcpy(xdname, name);
-			name = fname;
-
-			if (name[0] == '.' && name[1] == '/') {
-				for (name++; name[0] == '/'; name++)
-					/* LINTED */
-					;
-				xlen -= name - fname;
-			}
-			if (name[0] == '\0') {
-				name = ".";
-				xlen = 1;
-			}
+#if	defined(HAVE_FSTATAT) && defined(HAVE_DIRFD)
+			/*
+			 * We get the performance win only with the native
+			 * fstatat() and not from the emulation.
+			 */
+			if (d ? !getinfoat(dirfd(d), snm, ninfo) :
+				!getinfo(name, ninfo)) {
+#else
 			if (!getinfo(name, ninfo)) {
+#endif
 				if (!errhidden(E_STAT, name)) {
 					if (!errwarnonly(E_STAT, name))
 						xstats.s_staterrs++;
@@ -1641,6 +1684,19 @@ put_dir(sname, dname, namlen, info, ptb, last)
 				seterrno(0);
 				offset = telldir(d);
 				if (geterrno()) {
+					char	*p;
+
+					if (pathp->ps_tail == 0) {
+						dname = ".";
+						p = NULL;
+					} else {
+						/*
+						 * pathstore_t relocates ps_path
+						 */
+						dname = pathp->ps_path;
+						p = &pathp->ps_path[pathp->ps_tail-1];
+						*p = '\0';
+					}
 					if (!errhidden(E_OPEN, dname)) {
 						if (!errwarnonly(E_OPEN, dname))
 							xstats.s_openerrs++;
@@ -1650,6 +1706,8 @@ put_dir(sname, dname, namlen, info, ptb, last)
 						(void) errabort(E_OPEN, dname,
 									TRUE);
 					}
+					if (p)
+						*p = '/';
 					/*
 					 * XXX xstats.s_openerrs is wrong here.
 					 * Avoid an endless loop on unseekable
@@ -1661,10 +1719,26 @@ put_dir(sname, dname, namlen, info, ptb, last)
 				closedir(d);
 			}
 #endif
-			createi(name, name, xlen, ninfo, &thisd);
+			pathp->ps_tail += xlen;
+			createi(name, name, xlen+namlen, ninfo, pathp, &thisd);
+			pathp->ps_tail -= xlen;
+			pathp->ps_path[pathp->ps_tail] = '\0';
 #ifdef	HAVE_SEEKDIR
 			if (d && is_dir(ninfo) && depth <= 0) {
-				if (!(d = opendir(sname))) {
+				char	*p;
+
+				if (pathp->ps_tail == 0) {
+					sname = dname = ".";
+					p = NULL;
+				} else {
+					/*
+					 * pathstore_t relocates ps_path
+					 */
+					sname = dname = pathp->ps_path;
+					p = &pathp->ps_path[pathp->ps_tail-1];
+					*p = '\0';
+				}
+				if (!(d = lopendir(sname))) {
 					if (!errhidden(E_OPEN, dname)) {
 						if (!errwarnonly(E_OPEN, dname))
 							xstats.s_openerrs++;
@@ -1673,6 +1747,8 @@ put_dir(sname, dname, namlen, info, ptb, last)
 						(void) errabort(E_OPEN, dname,
 									TRUE);
 					}
+					if (p)
+						*p = '/';
 					break;
 				}
 				seterrno(0);
@@ -1692,24 +1768,51 @@ put_dir(sname, dname, namlen, info, ptb, last)
 					 * Avoid an endless loop on unseekable
 					 * directories.
 					 */
+					if (p)
+						*p = '/';
 					break;
 				}
+				if (p)
+					*p = '/';
 			}
 #endif
 		}
 	}
 out:
+	if (didslash)
+		pathp->ps_path[--pathp->ps_tail] = '\0';
 	if (d)
 		closedir(d);
 	if (dp)
 		free(info->f_dir);
 	if (info->f_dirinos)
 		free(info->f_dirinos);
+	if (pathp == &path)
+		free_pspace(pathp);
 	depth++;
 	if (!nodir && dirmode && putdir)
 		put_tcb(ptb, info);
 }
 
+/*
+ * This is currently only called in case we did run out of memory.
+ */
+LOCAL void
+toolong(dname, name, len)
+	char	*dname;
+	char	*name;
+	size_t	len;
+{
+	if (!errhidden(E_NAMETOOLONG, name)) {
+		if (!errwarnonly(E_NAMETOOLONG, name))
+			xstats.s_toolong++;
+		errmsgno(EX_BAD, "%s%s: Name too long (%lu), out of memory.\n",
+			dname, name,
+			(unsigned long)len);
+		(void) errabort(E_NAMETOOLONG, name, TRUE);
+	}
+}
+
 LOCAL BOOL
 checkdirexclude(sname, name, namlen, info)
 	char	*sname;
@@ -1719,40 +1822,60 @@ checkdirexclude(sname, name, namlen, info)
 {
 	FINFO	finfo;
 	char	pname[PATH_MAX+1];
+	char	*pnamep = pname;
+	size_t	pnamelen = sizeof (pname);
+	size_t	pnlen;
 	int	OFflag = Fflag;
 	char	*p;
+	BOOL	ret = TRUE;
 
 	Fflag = 0;
-	strcpy(pname, name);
-	p = &pname[namlen];
+	pnlen = namlen;
+	pnlen += 2 + 8;		/* Null Byte + '/' + strlen(".exclude") */
+	if (pnlen > pnamelen) {
+		pnamep = __fjmalloc(stderr, pnlen, "name buffer", JM_RETURN);
+		if (pnamep == NULL) {
+			errmsg("Cannot check '%s' for exclude\n", name);
+			goto notfound;
+		}
+	}
+	strcpy(pnamep, name);
+	p = &pnamep[namlen];
 	if (p[-1] != '/') {
 		*p++ = '/';
 	}
 	strcpy(p, ".mirror");
-	if (!_getinfo(pname, &finfo)) {
+	if (!_getinfo(pnamep, &finfo)) {
 		strcpy(p, ".exclude");
-		if (!_getinfo(pname, &finfo))
+		if (!_getinfo(pnamep, &finfo))
 			goto notfound;
 	}
 	if (is_file(&finfo)) {
 		if (OFflag == 3) {
 			nodesc++;
-			if (!dirmode)
-				createi(sname, name, namlen, info, (struct pdirs *)0);
-			create(pname, FALSE, FALSE);	/* Needed to strip off "./" */
-			if (dirmode)
-				createi(sname, name, namlen, info, (struct pdirs *)0);
+			if (!dirmode) {
+				createi(sname, name, namlen, info,
+					(pathstore_t *)0,
+					(struct pdirs *)0);
+			}
+			create(pnamep, FALSE, FALSE);	/* Needed to strip off "./" */
+			if (dirmode) {
+				createi(sname, name, namlen, info,
+					(pathstore_t *)0,
+					(struct pdirs *)0);
+			}
 			nodesc--;
 		}
-		Fflag = OFflag;
-		return (FALSE);
+		ret = FALSE;
 	}
 notfound:
+	if (pnamep != pname)
+		free(pnamep);
 	Fflag = OFflag;
-	return (TRUE);
+	return (ret);
 }
 
-EXPORT BOOL
+LOCAL BOOL
 checkexclude(sname, name, namlen, info)
 	char	*sname;
 	char	*name;
@@ -1842,8 +1965,9 @@ walkfunc(nm, fs, type, state)
 
 		finfo.f_sname = nm + state->base;
 		finfo.f_name = nm;
-		stat_to_info(fs, &finfo);
-		createi(nm + state->base, nm, strlen(nm), &finfo, (struct pdirs *)0);
+		stat_to_info(AT_FDCWD, fs, &finfo);
+		createi(nm + state->base, nm, strlen(nm), &finfo,
+			(pathstore_t *)0, (struct pdirs *)0);
 	}
 	return (0);
 }
