@@ -1,13 +1,13 @@
-/* @(#)device.c	1.15 09/07/11 Copyright 1996-2009 J. Schilling */
+/* @(#)device.c	1.17 18/06/16 Copyright 1996-2018 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)device.c	1.15 09/07/11 Copyright 1996-2009 J. Schilling";
+	"@(#)device.c	1.17 18/06/16 Copyright 1996-2018 J. Schilling";
 #endif
 /*
  *	Handle local and remote device major/minor mappings
  *
- *	Copyright (c) 1996-2009 J. Schilling
+ *	Copyright (c) 1996-2018 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -16,12 +16,16 @@ static	UConst char sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
  */
 
 #include <schily/standard.h>
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
 #include <schily/schily.h>
 #define	__XDEV__	/* Needed to activate XDEV_T definitions */
 #include <schily/device.h>
@@ -69,39 +73,39 @@ EXPORT	XDEV_T	_dev_mask[] = {
 	0x3FFFFFFFLU,
 	0x7FFFFFFFLU,
 	0xFFFFFFFFLU,
-#if SIZEOF_UNSIGNED_LONG_INT > 4
-	0x00000001FFFFFFFFLU,
-	0x00000003FFFFFFFFLU,
-	0x00000007FFFFFFFFLU,
-	0x0000000FFFFFFFFFLU,
-	0x0000001FFFFFFFFFLU,
-	0x0000003FFFFFFFFFLU,
-	0x0000007FFFFFFFFFLU,
-	0x000000FFFFFFFFFFLU,
-	0x000001FFFFFFFFFFLU,
-	0x000003FFFFFFFFFFLU,
-	0x000007FFFFFFFFFFLU,
-	0x00000FFFFFFFFFFFLU,
-	0x00001FFFFFFFFFFFLU,
-	0x00003FFFFFFFFFFFLU,
-	0x00007FFFFFFFFFFFLU,
-	0x0000FFFFFFFFFFFFLU,
-	0x0001FFFFFFFFFFFFLU,
-	0x0003FFFFFFFFFFFFLU,
-	0x0007FFFFFFFFFFFFLU,
-	0x000FFFFFFFFFFFFFLU,
-	0x001FFFFFFFFFFFFFLU,
-	0x003FFFFFFFFFFFFFLU,
-	0x007FFFFFFFFFFFFFLU,
-	0x00FFFFFFFFFFFFFFLU,
-	0x01FFFFFFFFFFFFFFLU,
-	0x03FFFFFFFFFFFFFFLU,
-	0x07FFFFFFFFFFFFFFLU,
-	0x0FFFFFFFFFFFFFFFLU,
-	0x1FFFFFFFFFFFFFFFLU,
-	0x3FFFFFFFFFFFFFFFLU,
-	0x7FFFFFFFFFFFFFFFLU,
-	0xFFFFFFFFFFFFFFFFLU,
+#if SIZEOF_ULLONG > 4
+	0x00000001FFFFFFFFLLU,
+	0x00000003FFFFFFFFLLU,
+	0x00000007FFFFFFFFLLU,
+	0x0000000FFFFFFFFFLLU,
+	0x0000001FFFFFFFFFLLU,
+	0x0000003FFFFFFFFFLLU,
+	0x0000007FFFFFFFFFLLU,
+	0x000000FFFFFFFFFFLLU,
+	0x000001FFFFFFFFFFLLU,
+	0x000003FFFFFFFFFFLLU,
+	0x000007FFFFFFFFFFLLU,
+	0x00000FFFFFFFFFFFLLU,
+	0x00001FFFFFFFFFFFLLU,
+	0x00003FFFFFFFFFFFLLU,
+	0x00007FFFFFFFFFFFLLU,
+	0x0000FFFFFFFFFFFFLLU,
+	0x0001FFFFFFFFFFFFLLU,
+	0x0003FFFFFFFFFFFFLLU,
+	0x0007FFFFFFFFFFFFLLU,
+	0x000FFFFFFFFFFFFFLLU,
+	0x001FFFFFFFFFFFFFLLU,
+	0x003FFFFFFFFFFFFFLLU,
+	0x007FFFFFFFFFFFFFLLU,
+	0x00FFFFFFFFFFFFFFLLU,
+	0x01FFFFFFFFFFFFFFLLU,
+	0x03FFFFFFFFFFFFFFLLU,
+	0x07FFFFFFFFFFFFFFLLU,
+	0x0FFFFFFFFFFFFFFFLLU,
+	0x1FFFFFFFFFFFFFFFLLU,
+	0x3FFFFFFFFFFFFFFFLLU,
+	0x7FFFFFFFFFFFFFFFLLU,
+	0xFFFFFFFFFFFFFFFFLLU,
 #endif
 };
 #else
@@ -139,39 +143,39 @@ EXPORT	XDEV_T	_dev_mask[] = {
 	0x3FFFFFFFL,
 	0x7FFFFFFFL,
 	0xFFFFFFFFL,
-#if SIZEOF_UNSIGNED_LONG_INT > 4
-	0x00000001FFFFFFFFL,
-	0x00000003FFFFFFFFL,
-	0x00000007FFFFFFFFL,
-	0x0000000FFFFFFFFFL,
-	0x0000001FFFFFFFFFL,
-	0x0000003FFFFFFFFFL,
-	0x0000007FFFFFFFFFL,
-	0x000000FFFFFFFFFFL,
-	0x000001FFFFFFFFFFL,
-	0x000003FFFFFFFFFFL,
-	0x000007FFFFFFFFFFL,
-	0x00000FFFFFFFFFFFL,
-	0x00001FFFFFFFFFFFL,
-	0x00003FFFFFFFFFFFL,
-	0x00007FFFFFFFFFFFL,
-	0x0000FFFFFFFFFFFFL,
-	0x0001FFFFFFFFFFFFL,
-	0x0003FFFFFFFFFFFFL,
-	0x0007FFFFFFFFFFFFL,
-	0x000FFFFFFFFFFFFFL,
-	0x001FFFFFFFFFFFFFL,
-	0x003FFFFFFFFFFFFFL,
-	0x007FFFFFFFFFFFFFL,
-	0x00FFFFFFFFFFFFFFL,
-	0x01FFFFFFFFFFFFFFL,
-	0x03FFFFFFFFFFFFFFL,
-	0x07FFFFFFFFFFFFFFL,
-	0x0FFFFFFFFFFFFFFFL,
-	0x1FFFFFFFFFFFFFFFL,
-	0x3FFFFFFFFFFFFFFFL,
-	0x7FFFFFFFFFFFFFFFL,
-	0xFFFFFFFFFFFFFFFFL,
+#if SIZEOF_ULLONG > 4
+	0x00000001FFFFFFFFLL,
+	0x00000003FFFFFFFFLL,
+	0x00000007FFFFFFFFLL,
+	0x0000000FFFFFFFFFLL,
+	0x0000001FFFFFFFFFLL,
+	0x0000003FFFFFFFFFLL,
+	0x0000007FFFFFFFFFLL,
+	0x000000FFFFFFFFFFLL,
+	0x000001FFFFFFFFFFLL,
+	0x000003FFFFFFFFFFLL,
+	0x000007FFFFFFFFFFLL,
+	0x00000FFFFFFFFFFFLL,
+	0x00001FFFFFFFFFFFLL,
+	0x00003FFFFFFFFFFFLL,
+	0x00007FFFFFFFFFFFLL,
+	0x0000FFFFFFFFFFFFLL,
+	0x0001FFFFFFFFFFFFLL,
+	0x0003FFFFFFFFFFFFLL,
+	0x0007FFFFFFFFFFFFLL,
+	0x000FFFFFFFFFFFFFLL,
+	0x001FFFFFFFFFFFFFLL,
+	0x003FFFFFFFFFFFFFLL,
+	0x007FFFFFFFFFFFFFLL,
+	0x00FFFFFFFFFFFFFFLL,
+	0x01FFFFFFFFFFFFFFLL,
+	0x03FFFFFFFFFFFFFFLL,
+	0x07FFFFFFFFFFFFFFLL,
+	0x0FFFFFFFFFFFFFFFLL,
+	0x1FFFFFFFFFFFFFFFLL,
+	0x3FFFFFFFFFFFFFFFLL,
+	0x7FFFFFFFFFFFFFFFLL,
+	0xFFFFFFFFFFFFFFFFLL,
 #endif
 };
 #endif
