@@ -1,13 +1,13 @@
-/* @(#)list.c	1.75 13/11/05 Copyright 1985, 1995, 2000-2013 J. Schilling */
+/* @(#)list.c	1.83 19/04/07 Copyright 1985, 1995, 2000-2019 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)list.c	1.75 13/11/05 Copyright 1985, 1995, 2000-2013 J. Schilling";
+	"@(#)list.c	1.83 19/04/07 Copyright 1985, 1995, 2000-2019 J. Schilling";
 #endif
 /*
  *	List the content of an archive
  *
- *	Copyright (c) 1985, 1995, 2000-2013 J. Schilling
+ *	Copyright (c) 1985, 1995, 2000-2019 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -31,6 +31,8 @@ static	UConst char sccsid[] =
 #include <schily/standard.h>
 #include <schily/stdlib.h>
 #include <schily/string.h>
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
 #include <schily/schily.h>
 #include "starsubs.h"
 #ifdef	USE_FIND
@@ -86,15 +88,22 @@ extern	struct WALK walkstate;
 		FINFO	newinfo;
 		TCB	tb;
 		TCB	newtb;
-		char	name[PATH_MAX+1];
-		char	lname[PATH_MAX+1];
-		char	newname[PATH_MAX+1];
-		char	newlname[PATH_MAX+1];
 	register TCB 	*ptb = &tb;
 
 	fillbytes((char *)&finfo, sizeof (finfo), '\0');
 	fillbytes((char *)&newinfo, sizeof (newinfo), '\0');
 
+	if (init_pspace(PS_STDERR, &finfo.f_pname) < 0)
+		return;
+	if (init_pspace(PS_STDERR, &finfo.f_plname) < 0)
+		return;
+	if (listnew || listnewf) {
+		if (init_pspace(PS_STDERR, &newinfo.f_pname) < 0)
+			return;
+		if (init_pspace(PS_STDERR, &newinfo.f_plname) < 0)
+			return;
+	}
+
 #ifdef	USE_FIND
 	if (dofind) {
 		walkopen(&walkstate);
@@ -108,8 +117,8 @@ extern	struct WALK walkstate;
 		if (prblockno)
 			(void) tblocks();		/* set curblockno */
 
-		finfo.f_name = name;
-		finfo.f_lname = lname;
+		finfo.f_name = finfo.f_pname.ps_path;
+		finfo.f_lname = finfo.f_plname.ps_path;
 		if (tcb_to_info(ptb, &finfo) == EOF)
 			break;
 		if (xdebug > 0)
@@ -128,7 +137,7 @@ extern	struct WALK walkstate;
 
 		if (listnew || listnewf) {
 			/*
-			 * XXX nsec beachten wenn im Archiv!
+			 * nsec beachten wenn im Archiv!
 			 */
 			if (((finfo.f_mtime > newinfo.f_mtime) ||
 			    ((finfo.f_xflags & XF_MTIME) &&
@@ -136,22 +145,26 @@ extern	struct WALK walkstate;
 			    (finfo.f_mtime == newinfo.f_mtime) &&
 			    (finfo.f_mnsec > newinfo.f_mnsec))) &&
 					(!listnewf || is_file(&finfo))) {
-				movebytes(&finfo, &newinfo, sizeof (finfo));
+				movebytes(&finfo, &newinfo,
+						offsetof(FINFO, f_pname));
 				movetcb(&tb, &newtb);
-				/*
-				 * Paranoia.....
-				 */
-				strncpy(newname, name, PATH_MAX);
-				newname[PATH_MAX] = '\0';
-				newinfo.f_name = newname;
+				if (strcpy_pspace(PS_STDERR,
+						&newinfo.f_pname,
+						finfo.f_name) < 0) {
+					newinfo.f_name = "";
+				} else {
+					newinfo.f_name =
+						newinfo.f_pname.ps_path;
+				}
 				if (newinfo.f_lname[0] != '\0') {
-					/*
-					 * Paranoia.....
-					 */
-					strncpy(newlname, newinfo.f_lname,
-								PATH_MAX);
-					newlname[PATH_MAX] = '\0';
-					newinfo.f_lname = newlname;
+					if (strcpy_pspace(PS_STDERR,
+							&newinfo.f_plname,
+							finfo.f_lname) < 0) {
+						newinfo.f_lname = "";
+					} else {
+						newinfo.f_lname =
+						    newinfo.f_plname.ps_path;
+					}
 				}
 				newinfo.f_flags |= F_HAS_NAME;
 			}
@@ -217,6 +230,7 @@ modstr(info, s, mode)
 	register char	*str = s;
 	register int	i;
 
+	*str++ = '?';				/* Unknown file type */
 	for (i = 9; --i >= 0; ) {
 		if (mode & (1 << i))
 			*str++ = mstr[i];
@@ -233,28 +247,28 @@ modstr(info, s, mode)
 	str = s;
 	if (mode & TSVTX) {
 		if (mode & TOEXEC) {
-			str[8] = 't';		/* Sticky & exec. by others  */
+			str[9] = 't';		/* Sticky & exec. by others  */
 		} else {
-			str[8] = 'T';		/* Sticky but !exec. by oth  */
+			str[9] = 'T';		/* Sticky but !exec. by oth  */
 		}
 	}
 	if (mode & TSGID) {
 		if (mode & TGEXEC) {
-			str[5] = 's';		/* Sgid & executable by grp  */
+			str[6] = 's';		/* Sgid & executable by grp  */
 		} else {
-			if (is_dir(info))
-				str[5] = 'S';	/* Sgid directory	    */
+			if (!is_file(info))
+				str[6] = 'S';	/* Sgid directory, or other  */
 			else
-				str[5] = 'l';	/* Mandatory lock file	    */
+				str[6] = 'l';	/* Mandatory lock file	    */
 		}
 	}
 	if (mode & TSUID) {
 		if (mode & TUEXEC)
-			str[2] = 's';		/* Suid & executable by own. */
+			str[3] = 's';		/* Suid & executable by own. */
 		else
-			str[2] = 'S';		/* Suid but not executable   */
+			str[3] = 'S';		/* Suid but not executable   */
 	}
-	i = 9;
+	i = 10;
 #ifdef	USE_ACL
 	if ((info->f_xflags & (XF_ACL_ACCESS|XF_ACL_DEFAULT|XF_ACL_ACE)) != 0)
 		str[i++] = '+';
@@ -273,15 +287,15 @@ list_file(info)
 		FILE	*f;
 		time_t	*tp;
 		char	*tstr;
-		char	mstr[12]; /* 9 UNIX chars + ACL '+' XATTR '@' + nul */
-		char	lstr[11]; /* contains link count as string */
-	static	char	nuid[11]; /* XXXX 64 bit longs??? */
-	static	char	ngid[11]; /* XXXX 64 bit longs??? */
+		char	mstr[13]; /* 10 UNIX chars + ACL '+' XATTR '@' + nul */
+		char	lstr[22]; /* ' ' + link count as string - 64 bits */
+	static	char	nuid[21]; /* uid as 64 bit long */
+	static	char	ngid[21]; /* gid as 64 bit long */
 		char	*add = "";
 
 	f = vpr;
 	if (prblockno)
-		fprintf(f, "block %9lld: ", curblockno);
+		fgtprintf(f, "block %9lld: ", curblockno);
 	if (cflag)
 		fprintf(f, "a ");
 	else if (xflag)
@@ -345,14 +359,15 @@ if (xft == 0 || xft == XT_BAD) {
 }
 		if (xft == XT_LINK)
 			xft = info->f_rxftype;
+		{
+			char	*p = XTTOSTR(xft);
+
+			if (p)
+				mstr[0] = *p;
+		}
 		if (!paxls) {
 			fprintf(f,
-				" %s%s%s %3.*s/%-3.*s %.12s %4.4s ",
-#ifdef	OLD
-				typetab[info->f_filetype & 07],
-#else
-				XTTOSTR(xft),
-#endif
+				" %s%s %3.*s/%-3.*s %.12s %4.4s ",
 				mstr,
 				lstr,
 				(int)info->f_umaxlen, info->f_uname,
@@ -360,12 +375,7 @@ if (xft == 0 || xft == XT_BAD) {
 				&tstr[4], &tstr[20]);
 		} else {
 			fprintf(f,
-				"%s%s%s %-8.*s %-8.*s ",
-#ifdef	OLD
-				typetab[info->f_filetype & 07],
-#else
-				XTTOSTR(xft),
-#endif
+				"%s%s %-8.*s %-8.*s ",
 				mstr,
 				lstr,
 				(int)info->f_umaxlen, info->f_uname,
@@ -389,19 +399,23 @@ if (xft == 0 || xft == XT_BAD) {
 		fprintf(f, "\n");
 		return;
 	}
+	/*
+	 * In case of a hardlinked symlink, we currently do not have the symlink
+	 * target path and thus cannot check the synlink target. So first check
+	 * whether it is a hardlink.
+	 */
 	if (is_link(info)) {
 		if (is_dir(info))
-			fprintf(f, " directory");
+			fgtprintf(f, " directory");
 		fprintf(f, " %s %s",
 			paxls ? "==" : "link to",
 			info->f_lname);
-	}
-	if (is_symlink(info))
+	} else if (is_symlink(info))
 		fprintf(f, " -> %s", info->f_lname);
 	if (is_volhdr(info))
-		fprintf(f, " --Volume Header--");
+		fgtprintf(f, " --Volume Header--");
 	if (is_multivol(info)) {
-		fprintf(f, " --Continued at byte %lld--",
+		fgtprintf(f, " --Continued at byte %lld--",
 						(Llong)info->f_contoffset);
 	}
 	fprintf(f, "\n");
@@ -424,7 +438,7 @@ vprint(info)
 		f = vpr;
 
 		if (prblockno)
-			fprintf(f, "block %9lld: ", curblockno);
+			fgtprintf(f, "block %9lld: ", curblockno);
 		if (cflag)
 			mode = "a ";
 		else if (xflag)
@@ -446,12 +460,12 @@ vprint(info)
 		}
 		if (is_dir(info)) {
 			if (is_link(info)) {
-				fprintf(f, "%s%s%s directory %s %s\n",
+				fgtprintf(f, "%s%s%s directory %s %s\n",
 					mode, info->f_name, add,
 					paxls ? "==" : "link to",
 					info->f_lname);
 			} else {
-				fprintf(f, "%s%s%s directory\n", mode,
+				fgtprintf(f, "%s%s%s directory\n", mode,
 							info->f_name, add);
 			}
 		} else if (is_link(info)) {
@@ -465,9 +479,9 @@ vprint(info)
 				paxls ? "->" : "symbolic link to",
 				info->f_lname);
 		} else if (is_special(info)) {
-			fprintf(f, "%s%s special\n", mode, info->f_name);
+			fgtprintf(f, "%s%s special\n", mode, info->f_name);
 		} else {
-			fprintf(f, "%s%s %lld bytes, %lld tape blocks\n",
+			fgtprintf(f, "%s%s %lld bytes, %lld tape blocks\n",
 				mode, info->f_name, (Llong)info->f_size,
 				(Llong)tarblocks(info->f_rsize));
 		}
