@@ -1,12 +1,13 @@
-/* @(#)pax.c	1.24 07/10/20 Copyright 1989, 2003-2007 J. Schilling */
+/* @(#)pax.c	1.41 19/03/27 Copyright 1989, 2003-2019 J. Schilling */
+#include <schily/mconfig.h>
 #ifndef lint
-static	char _p_sccsid[] =
-	"@(#)pax.c	1.24 07/10/20 Copyright 1989, 2003-2007 J. Schilling";
+static	const char _p_sccsid[] =
+	"@(#)pax.c	1.41 19/03/27 Copyright 1989, 2003-2019 J. Schilling";
 #endif
 /*
  *	PAX specific routines for star main program.
  *
- *	Copyright (c) 1989, 2003-2007 J. Schilling
+ *	Copyright (c) 1989, 2003-2019 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -15,6 +16,8 @@ static	char _p_sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
@@ -81,12 +84,15 @@ LOCAL	void	pax_setopts	__PR((char *o));
 /*
  * PAX related options
  *
- * The official POSIX options start after the -bz/-lzo/-7z option.
+ * The official POSIX options start after the -bz/-lzo/-7z/-xz/-lzip/-zstd option.
  */
 /* BEGIN CSTYLED */
-char	_opts[] = "help,xhelp,version,debug,xdebug#,xd#,time,no-statistics,do-statistics,fifostats,numeric,no-fifo,no-fsync,bs&,fs&,/,..,secure-links,acl,xfflags,z,bz,lzo,7z,r,w,a,b&,c,d,f&,H,i,k,L,l,n,o*,p&,s&,t,u,v+,x&,artype&,X";
+char	_opts[] = "help,xhelp,version,debug,xdebug#,xd#,time,no-statistics,do-statistics,fifostats,numeric,no-fifo,no-fsync,do-fsync%0,bs&,fs&,/,..,secure-links,no-secure-links%0,acl,xfflags,z,bz,lzo,7z,xz,lzip,zstd,r,w,a,b&,c,d,f&,H,i,k,L,l,n,o*,p&,s&,t,u,v+,x&,artype&,X";
 /* END CSTYLED */
 char	*opts = _opts;
+#ifdef	NO_STAR_MAIN
+struct ga_props	gaprops;
+#endif
 
 LOCAL	void	pax_info	__PR((void));
 
@@ -119,6 +125,7 @@ gargs(ac, av)
 #ifdef	STAR_MAIN
 	pax_setopts(opts);			/* set up opts for getfiles */
 #endif
+	getarginit(&gaprops, GAF_SINGLEARG);	/* POSIX combined args	  */
 
 	iftype		= I_PAX;		/* command line interface */
 	ptype		= P_PAX;		/* program interface type */
@@ -126,23 +133,29 @@ gargs(ac, av)
 	paxmatch	= TRUE;
 	nopflag		= TRUE;			/* pax default */
 	no_stats	= TRUE;			/* -no-statitstics	   */
+	nochown		= TRUE;			/* chown only with -po / -pe */
 
+	if (pname) {				/* cli=xxx seen as argv[1] */
+		--ac, av++;
+	}
 	--ac, ++av;
 	files = getfilecount(ac, av, opts);
-	if (getallargs(&ac, &av, opts,
+	if (getlallargs(&ac, &av, &gaprops, opts,
 				&help, &xhelp, &prvers, &debug, &xdebug, &xdebug,
 #ifndef	__old__lint
 				&showtime, &no_stats, & do_stats, &do_fifostats,
-				&numeric,  &no_fifo, &no_fsync,
-				getnum, &bs,
-				getnum, &fs,
-				&abs_path, &allow_dotdot, &secure_links,
+				&numeric,  &no_fifo, &no_fsync, &no_fsync,
+				getenum, &bs,
+				getenum, &fs,
+				&abs_path, &allow_dotdot,
+				&secure_links, &secure_links,
 				&doacl, &dofflags,
-				&zflag, &bzflag, &lzoflag, &p7zflag,
+				&zflag, &bzflag, &lzoflag,
+				&p7zflag, &xzflag, &lzipflag, &zstdflag,
 				&paxrflag,
 				&paxwflag,
 				&paxaflag,
-				getnum, &bs,
+				getenum, &bs,
 				&notarg,		/* -c */
 				&paxdflag,		/* -d */
 				addtarfile, NULL,	/* -f */
@@ -153,7 +166,7 @@ gargs(ac, av)
 				&paxlflag,		/* -l */
 				&paxnflag,		/* -n */
 				&paxopts,		/* -o */
-				getpriv, NULL,		/* -p */
+				getpaxpriv, NULL,	/* -p */
 				parsesubst, &do_subst,	/* -s */
 				&paxtflag,		/* -t */
 				&paxuflag,		/* -u */
@@ -161,7 +174,7 @@ gargs(ac, av)
 				gethdr, &chdrtype,	/* -x */
 				gethdr, &chdrtype,	/* artype= */
 #endif /* __old__lint */
-				nomount) < 0) {
+				&nomount) < 0) {
 		errmsgno(EX_BAD, "Bad Option: %s.\n", av[0]);
 		susage(EX_BAD);
 	}
@@ -202,11 +215,6 @@ gargs(ac, av)
 		errmsgno(EX_BAD, "Unsupported option -l.\n");
 		susage(EX_BAD);
 	}
-	if (paxopts) {
-		/* Unsupported in UNIX-98 */
-		errmsgno(EX_BAD, "SUSv2 / UNIX-98 does not specify any option for -o.\n");
-		susage(EX_BAD);
-	}
 	uncond = !paxuflag;
 
 	if (do_stats)
@@ -214,7 +222,9 @@ gargs(ac, av)
 
 	star_checkopts(/* oldtar */ FALSE, /* dodesc */ TRUE,
 				/* usetape */ FALSE,
-				/* archive */ -1, no_fifo, /* llbs */ 0);
+				/* archive */ -1, no_fifo,
+				paxopts,
+				/* llbs */ 0);
 
 	nolinkerr = FALSE;
 }
@@ -222,8 +232,10 @@ gargs(ac, av)
 LOCAL void
 pax_info()
 {
+	const	char	*n = pname ? pname : get_progname();
+
 	error("\nFor a more complete user interface use the tar type command interface.\n");
-	error("See 'man star'. The %s command is more or less limited to the\n", get_progname());
+	error("See 'man star'. The %s command is more or less limited to the\n", n);
 	error("POSIX standard pax command line interface.\n");
 }
 
@@ -234,11 +246,13 @@ LOCAL void
 susage(ret)
 	int	ret;
 {
-	error("Usage:\t%s cmd [options] file1 ... filen\n", get_progname());
-	error("\nUse\t%s -help\n", get_progname());
-	error("and\t%s -xhelp\n", get_progname());
+	const	char	*n = pname ? pname : get_progname();
+
+	error("Usage:\t%s cmd [options] file1 ... filen\n", n);
+	error("\nUse\t%s -help\n", n);
+	error("and\t%s -xhelp\n", n);
 	error("to get a list of valid cmds and options.\n");
-	error("\nUse\t%s -x help\n", get_progname());
+	error("\nUse\t%s -x help\n", n);
 	error("to get a list of valid archive header formats.\n");
 	pax_info();
 	exit(ret);
@@ -249,7 +263,9 @@ LOCAL void
 usage(ret)
 	int	ret;
 {
-	error("Usage:\t%s cmd [options] file1 ... filen\n", get_progname());
+	const	char	*n = pname ? pname : get_progname();
+
+	error("Usage:\t%s cmd [options] file1 ... filen\n", n);
 	error("Cmd:\n");
 	error("\t<none>\t\tlist named files from tape\n");
 	error("\t-r\t\textract named files from tape\n");
@@ -264,14 +280,16 @@ usage(ret)
 	error("\t-c\t\tinvert matching rules\n");
 	error("\t-d\t\tdo not descend directories\n");
 	error("\t-f nm\t\tuse 'nm' as tape instead of stdin/stdout\n");
+	error("\t-H\t\tfollow symbolic links from cmdline as if they were files\n");
 	error("\t-i\t\tdo interactive creation/extraction/renaming\n");
 	error("\t-k\t\tkeep existing files\n");
 	error("\t-l\t\tlink files rather than copying them\n");
+	error("\t-L\t\tfollow symbolic links as if they were files\n");
 	error("\t-n\t\tone match per pattern only\n");
 	error("\t-o\t\toptions (none specified with SUSv2 / UNIX-98)\n");
 	error("\t-p string\tset privileges\n");
 	error("\t-s replstr\tApply ed like pattern substitution -s /old/new/gp on filenames\n");
-	error("\t-t\t\tTIME\n");
+	error("\t-t\t\trestore atime after reading files\n");
 	error("\t-u\t\treplace/restore files only if they are newer\n");
 	error("\t-v\t\tincrement verbose level\n");
 	error("\t-x header\tgenerate 'header' type archive (see -x help)\n");
@@ -281,6 +299,9 @@ usage(ret)
 	error("\t-bz\t\t(*) pipe input/output through bzip2, does not work on tapes\n");
 	error("\t-lzo\t\t(*) pipe input/output through lzop, does not work on tapes\n");
 	error("\t-7z\t\t(*) pipe input/output through p7zip, does not work on tapes\n");
+	error("\t-xz\t\t(*) pipe input/output through xz, does not work on tapes\n");
+	error("\t-lzip\t\t(*) pipe input/output through lzip, does not work on tapes\n");
+	error("\t-zstd\t\t(*) pipe input/output through zstd, does not work on tapes\n");
 #ifdef	FIFO
 	error("\t-no-fifo\t(*) don't use a fifo to optimize data flow from/to tape\n");
 #endif
@@ -294,13 +315,16 @@ LOCAL void
 xusage(ret)
 	int	ret;
 {
-	error("Usage:\t%s cmd [options] file1 ... filen\n", get_progname());
+	const	char	*n = pname ? pname : get_progname();
+
+	error("Usage:\t%s cmd [options] file1 ... filen\n", n);
 	error("Extended options:\n");
 	error("\t-debug\t\tprint additional debug messages\n");
 	error("\txdebug=#,xd=#\tset extended debug level\n");
 	error("\t-/\t\tdon't strip leading '/'s from file names\n");
 	error("\t-..\t\tdon't skip filenames that contain '..' in non-interactive extract\n");
-	error("\t-secure-links\tdon't extract links that start with '/' or contain '..'\n");
+	error("\t-secure-links\tdon't extract links that start with '/' or contain '..' (default)\n");
+	error("\t-no-secure-links\textract links that start with '/' or contain '..'\n");
 	error("\t-acl\t\thandle access control lists\n");
 	error("\t-xfflags\thandle extended file flags\n");
 	error("\tbs=#\t\tset (output) block size to #\n");
@@ -308,6 +332,7 @@ xusage(ret)
 	error("\tfs=#\t\tset fifo size to #\n");
 #endif
 	error("\t-no-fsync\tdo not call fsync() for each extracted file (may be dangerous)\n");
+	error("\t-do-fsync\tcall fsync() for each extracted file\n");
 	error("\t-time\t\tprint timing info\n");
 	error("\t-no-statistics\tdo not print statistics\n");
 	error("\t-do-statistics\tprint statistics\n");
