@@ -1,13 +1,13 @@
-/* @(#)remove.c	1.55 10/08/23 Copyright 1985, 1991-2010 J. Schilling */
+/* @(#)remove.c	1.61 18/10/23 Copyright 1985, 1991-2018 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)remove.c	1.55 10/08/23 Copyright 1985, 1991-2010 J. Schilling";
+	"@(#)remove.c	1.61 18/10/23 Copyright 1985, 1991-2018 J. Schilling";
 #endif
 /*
  *	remove files an file trees
  *
- *	Copyright (c) 1985, 1992-2010 J. Schilling
+ *	Copyright (c) 1985, 1992-2018 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -16,6 +16,8 @@ static	UConst char sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
@@ -23,14 +25,18 @@ static	UConst char sccsid[] =
 
 #include <schily/stdio.h>
 #include <schily/standard.h>
+#include <schily/fcntl.h>	/* Wegen AT_REMOVEDIR */
 #include "star.h"
 #include "table.h"
 #include <schily/dirent.h>	/* XXX Wegen S_IFLNK */
 #include <schily/unistd.h>
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
 #include <schily/string.h>
 #include <schily/schily.h>
 #include <schily/errno.h>
 #include "starsubs.h"
+#include "pathname.h"
 
 
 extern	FILE	*tty;
@@ -43,8 +49,10 @@ extern	BOOL	remove_recursive;
 extern	BOOL	keep_nonempty_dirs;
 
 EXPORT	BOOL	remove_file	__PR((char *name, BOOL isfirst));
-LOCAL	BOOL	_remove_file	__PR((char *name, BOOL isfirst, int depth));
-LOCAL	BOOL	remove_tree	__PR((char *name, BOOL isfirst, int depth));
+LOCAL	BOOL	_remove_file	__PR((char *name, pathstore_t *path,
+						BOOL isfirst, int depth));
+LOCAL	BOOL	remove_tree	__PR((char *name, pathstore_t *path,
+						BOOL isfirst, int depth));
 
 /*
  * Remove a file or a directory tree.
@@ -55,8 +63,11 @@ remove_file(name, isfirst)
 	register char	*name;
 		BOOL	isfirst;
 {
-	static	int	depth	= -10;
-	static	int	dinit	= 0;
+	static	int		depth	= -10;
+	static	int		dinit	= 0;
+		pathstore_t	path;
+		char		pbuf[PATH_MAX+1];
+		BOOL		ret;
 
 	if (!dinit) {
 #ifdef	_SC_OPEN_MAX
@@ -66,18 +77,26 @@ remove_file(name, isfirst)
 #endif
 		dinit = 1;
 	}
-	return (_remove_file(name, isfirst, depth));
+	path.ps_size = 0;
+	path.ps_tail = 0;
+	path.ps_path = pbuf;
+	ret = _remove_file(name, &path, isfirst, depth);
+	if (path.ps_path != pbuf)
+		free_pspace(&path);
+	return (ret);
 }
 
 LOCAL BOOL
-_remove_file(name, isfirst, depth)
-	register char	*name;
-		BOOL	isfirst;
-		int	depth;
+_remove_file(name, path, isfirst, depth)
+	register char		*name;
+		pathstore_t	*path;
+		BOOL		isfirst;
+		int		depth;
 {
 	char	buf[32];
 	char	ans = '\0';
 	int	err = EX_BAD;
+	int	len;
 	BOOL	fr_save = force_remove;
 	BOOL	rr_save = remove_recursive;
 	BOOL	ret;
@@ -85,8 +104,15 @@ _remove_file(name, isfirst, depth)
 	if (remove_first && !isfirst)
 		return (FALSE);
 	if (!force_remove && (interactive || ask_remove)) {
-		fprintf(vpr, "remove '%s' ? Y(es)/N(o) :", name); fflush(vpr);
-		fgetline(tty, buf, 2);
+		fgtprintf(vpr, "remove '%s' ? Y(es)/N(o) :", name); fflush(vpr);
+		buf[0] = '\0';
+		len = fgetstr(tty, buf, 3);
+		if (len > 0 && buf[len-1] != '\n') {
+			while (getc(tty) != '\n') {
+				if (feof(tty) || ferror(tty))
+					break;
+			}
+		}
 	}
 	if (force_remove ||
 	    ((interactive || ask_remove) && (ans = toupper(buf[0])) == 'Y')) {
@@ -94,10 +120,10 @@ _remove_file(name, isfirst, depth)
 		/*
 		 * only unlink non directories or empty directories
 		 */
-		if (rmdir(name) < 0) {
+		if (lunlinkat(name, AT_REMOVEDIR) < 0) {	/* rmdir() */
 			err = geterrno();
 			if (err == ENOTDIR) {
-				if (unlink(name) < 0) {
+				if (lunlinkat(name, 0) < 0) {
 					err = geterrno();
 #ifdef	RM_DEBUG
 					errmsg("rmdir: Not a dir but cannot unlink file.\n");
@@ -113,11 +139,18 @@ _remove_file(name, isfirst, depth)
 #endif
 				if (!remove_recursive) {
 					if (ans == 'Y') {
-						fprintf(vpr,
+						fgtprintf(vpr,
 						"Recursive remove nonempty '%s' ? Y(es)/N(o) :",
 							name);
 						fflush(vpr);
-						fgetline(tty, buf, 2);
+						buf[0] = '\0';
+						len = fgetstr(tty, buf, 3);
+						if (len > 0 && buf[len-1] != '\n') {
+							while (getc(tty) != '\n') {
+								if (feof(tty) || ferror(tty))
+									break;
+							}
+						}
 						if (toupper(buf[0]) == 'Y') {
 							force_remove = TRUE;
 							remove_recursive = TRUE;
@@ -133,7 +166,7 @@ _remove_file(name, isfirst, depth)
 						return (FALSE);
 					}
 				}
-				ret = remove_tree(name, isfirst, depth);
+				ret = remove_tree(name, path, isfirst, depth);
 
 				force_remove = fr_save;
 				remove_recursive = rr_save;
@@ -162,25 +195,41 @@ cannot:
 }
 
 LOCAL BOOL
-remove_tree(name, isfirst, depth)
+remove_tree(name, path, isfirst, depth)
 	register char	*name;
+		pathstore_t	*path;
 		BOOL	isfirst;
 		int	depth;
 {
 	DIR		*d;
 	struct dirent	*dir;
 	BOOL		ret = TRUE;
-	char		xn[PATH_MAX];	/* XXX A bad idea for a final solution */
+	size_t		otail;
+	size_t		nlen;
 	char		*p;
 
-	if ((d = opendir(name)) == NULL) {
+	if ((d = lopendir(name)) == NULL) {
 		return (FALSE);
 	}
 	depth--;
 
-	strcpy(xn, name);
-	p = &xn[strlen(name)];
-	*p++ = '/';
+	if (path->ps_tail == 0) {
+		nlen = strlen(name);
+		if (path->ps_size == 0 && nlen > (PATH_MAX-2)) {
+			/*
+			 * Does not fit into static buffer.
+			 */
+			init_pspace(PS_STDERR, path);
+			strcpy_pspace(PS_STDERR, path, name);
+		} else {
+			strcpy(path->ps_path, name);
+			path->ps_tail = nlen;
+		}
+	}
+	otail = path->ps_tail;
+	p = path->ps_path + path->ps_tail;
+	*p++ = '/';					/* Trailing '/' */
+	*p = '\0';
 
 	while ((dir = readdir(d)) != NULL) {
 
@@ -188,24 +237,45 @@ remove_tree(name, isfirst, depth)
 				streql(dir->d_name, ".."))
 			continue;
 
+		nlen = strlen(dir->d_name);
+		if ((nlen + 2 + path->ps_tail) > PATH_MAX) {
+			if (path->ps_size == 0) {
+				/*
+				 * Does not fit into static buffer.
+				 */
+				name[path->ps_tail + 1] = '\0';
+				init_pspace(PS_STDERR, path);
+				strcpy_pspace(PS_STDERR, path, name);
+				path->ps_tail--;	/* Trailing '/' */
+			}
+			grow_pspace(PS_STDERR,
+					path, (nlen + 2 + path->ps_tail));
+			p = path->ps_path + path->ps_tail + 1;
+			*p = 0;
+		}
 		strcpy(p, dir->d_name);
+		path->ps_tail += nlen + 1;
 
 		if (depth <= 0) {
 			closedir(d);
 		}
-		if (!_remove_file(xn, isfirst, depth))
+		if (!_remove_file(path->ps_path, path, isfirst, depth))
 			ret = FALSE;
-		if (depth <= 0 && (d = opendir(name)) == NULL) {
+		path->ps_tail = otail;
+
+		if (depth <= 0 && (d = lopendir(name)) == NULL) {
+			p[-1] = '\0';			/* Trailing '/' */
 			return (FALSE);
 		}
 	}
 
 	closedir(d);
+	p[-1] = '\0';					/* Trailing '/' */
 
 	if (ret == FALSE)
 		return (ret);
 
-	if (rmdir(name) >= 0)
+	if (lunlinkat(name, AT_REMOVEDIR) >= 0)	/* rmdir() */
 		return (ret);
 
 	errmsg("Directory '%s' not removed.\n", name);
