@@ -1,13 +1,13 @@
-/* @(#)xattr.c	1.13 09/07/11 Copyright 2003-2009 J. Schilling */
+/* @(#)xattr.c	1.22 18/08/06 Copyright 2003-2018 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)xattr.c	1.13 09/07/11 Copyright 2003-2009 J. Schilling";
+	"@(#)xattr.c	1.22 18/08/06 Copyright 2003-2018 J. Schilling";
 #endif
 /*
  *	Handle Extended File Attributes on Linux
  *
- *	Copyright (c) 2003-2009 J. Schilling
+ *	Copyright (c) 2003-2018 J. Schilling
  *	Thanks to Anreas Grnbacher <agruen@suse.de> for the
  *	first implemenation.
  */
@@ -18,6 +18,8 @@ static	UConst char sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
@@ -26,12 +28,18 @@ static	UConst char sccsid[] =
 #include <schily/stdio.h>
 #include <schily/stdlib.h>
 #include <schily/string.h>
+#if defined(HAVE_SYS_XATTR_H)
+#include <sys/xattr.h>
+#else
 #if defined(HAVE_ATTR_XATTR_H)
 #include <attr/xattr.h>
-#endif
+#endif	/* HAVE_ATTR_XATTR_H */
+#endif	/* HAVE_SYS_XATTR_H */
 #include "star.h"
 #include <schily/standard.h>
 #include <schily/unistd.h>
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
 #include <schily/schily.h>
 #include "starsubs.h"
 #include "checkerr.h"
@@ -47,6 +55,26 @@ static	UConst char sccsid[] =
 LOCAL star_xattr_t	*static_xattr;
 #endif
 
+#ifdef	USE_SELINUX
+extern	BOOL		selinux_enabled;
+#endif
+
+#ifdef	XATTR_NOFOLLOW		/* Max OS X has incompatible prototypes */
+
+/*
+ * This works since we currently only use lgetxattr()/lsetxattr()/llistxattr()
+ * but not getxattr()/setxattr()/listxattr(). If we ever need the latter,
+ * we would need to use a different macro names to be able to abstract from
+ * the differences between Linux and Mac OS X.
+ */
+#define	lgetxattr(p, n, v, s)	getxattr(p, n, v, s, 0, XATTR_NOFOLLOW)
+#define	lsetxattr(p, n, v, s, f) setxattr(p, n, v, s, 0, (f)|XATTR_NOFOLLOW)
+#define	llistxattr(p, nb, s)	listxattr(p, nb, s, XATTR_NOFOLLOW)
+
+#else	/* !XATTR_NOFOLLOW */
+/*
+ * This is for Linux
+ */
 #if defined(HAVE_GETXATTR) && !defined(HAVE_LGETXATTR)
 #define	lgetxattr	getxattr
 #endif
@@ -56,6 +84,7 @@ LOCAL star_xattr_t	*static_xattr;
 #if defined(HAVE_LISTXATTR) && !defined(HAVE_LLISTXATTR)
 #define	llistxattr	listxattr
 #endif
+#endif	/* !XATTR_NOFOLLOW */
 
 EXPORT void
 opt_xattr()
@@ -69,6 +98,14 @@ opt_xattr()
 #endif
 }
 
+EXPORT void
+opt_selinux()
+{
+#ifdef	USE_SELINUX
+	printf(" SELinux");
+#endif
+}
+
 /* ARGSUSED */
 EXPORT BOOL
 get_xattr(info)
@@ -88,7 +125,7 @@ get_xattr(info)
 	info->f_xflags &= ~XF_XATTR;
 	info->f_xattr = NULL;
 
-	list_len = llistxattr(info->f_name, NULL, 0);
+	list_len = llistxattr(info->f_sname, NULL, 0);
 	if (list_len < 0) {
 		if (!errhidden(E_GETXATTR, info->f_name)) {
 			if (!errwarnonly(E_GETXATTR, info->f_name))
@@ -101,7 +138,7 @@ get_xattr(info)
 		return (FALSE);
 	}
 	alist = ___malloc(list_len+2, "extended attribute");
-	list_len = llistxattr(info->f_name, alist, list_len);
+	list_len = llistxattr(info->f_sname, alist, list_len);
 	if (list_len < 0) {
 		if (!errhidden(E_GETXATTR, info->f_name)) {
 			if (!errwarnonly(E_GETXATTR, info->f_name))
@@ -146,7 +183,7 @@ get_xattr(info)
 		static_xattr[i].value = NULL;
 		strcpy(static_xattr[i].name, lp);
 
-		len = lgetxattr(info->f_name, lp, NULL, 0);
+		len = lgetxattr(info->f_sname, lp, NULL, 0);
 		if (len < 0) {
 			if (!errhidden(E_GETXATTR, info->f_name)) {
 				if (!errwarnonly(E_GETXATTR, info->f_name))
@@ -160,7 +197,7 @@ get_xattr(info)
 		}
 		static_xattr[i].value_len = len;
 		static_xattr[i].value = ___malloc(len, "extended attribute");
-		len = lgetxattr(info->f_name, lp, static_xattr[i].value, len);
+		len = lgetxattr(info->f_sname, lp, static_xattr[i].value, len);
 		if (len < 0) {
 			if (!errhidden(E_GETXATTR, info->f_name)) {
 				if (!errwarnonly(E_GETXATTR, info->f_name))
@@ -209,6 +246,11 @@ set_xattr(info)
 		return (TRUE);
 
 	for (xap = info->f_xattr; xap->name != NULL; xap++) {
+#ifdef	USE_SELINUX
+		if (selinux_enabled &&
+		    (strcmp(xap->name, "security.selinux") == 0))
+			continue;
+#endif
 		if (lsetxattr(info->f_name, xap->name, xap->value,
 		    xap->value_len, 0) != 0) {
 			if (!errhidden(E_SETXATTR, info->f_name)) {
@@ -226,6 +268,39 @@ set_xattr(info)
 	return (TRUE);
 }
 
+#ifdef	USE_SELINUX
+EXPORT BOOL
+setselinux(info)
+	register FINFO	*info;
+{
+	star_xattr_t	*xap;
+
+	if (info->f_xattr == NULL || (info->f_xflags & XF_XATTR) == 0) {
+		if (setfscreatecon(NULL) < 0)
+			goto err;
+		return (TRUE);
+	}
+
+	for (xap = info->f_xattr; xap->name != NULL; xap++) {
+		if (strcmp(xap->name, "security.selinux") == 0) {
+			if (setfscreatecon(xap->value) < 0)
+				goto err;
+			return (TRUE);
+		}
+	}
+	/*
+	 * There was no "security.selinux" label, so we need to clear
+	 * the context.
+	 */
+	if (setfscreatecon(NULL) < 0)
+		goto err;
+	return (TRUE);
+err:
+	errmsg("Cannot setup security context for '%s'.\n", info->f_name);
+	return (FALSE);
+}
+#endif	/* USE_SELINUX */
+
 EXPORT void
 free_xattr(xattr)
 	star_xattr_t	**xattr;
