@@ -1,14 +1,14 @@
-/* @(#)xheader.c	1.85 11/08/03 Copyright 2001-2011 J. Schilling */
+/* @(#)xheader.c	1.99 19/03/09 Copyright 2001-2019 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)xheader.c	1.85 11/08/03 Copyright 2001-2011 J. Schilling";
+	"@(#)xheader.c	1.99 19/03/09 Copyright 2001-2019 J. Schilling";
 #endif
 /*
  *	Handling routines to read/write, parse/create
  *	POSIX.1-2001 extended archive headers
  *
- *	Copyright (c) 2001-2011 J. Schilling
+ *	Copyright (c) 2001-2019 J. Schilling
  */
 /*
  * The contents of this file are subject to the terms of the
@@ -17,6 +17,8 @@ static	UConst char sccsid[] =
  * with the License.
  *
  * See the file CDDL.Schily.txt in this distribution for details.
+ * A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  *
  * When distributing Covered Code, include this CDDL HEADER in each
  * file and include the License file CDDL.Schily.txt from this distribution.
@@ -34,14 +36,18 @@ static	UConst char sccsid[] =
 #include <schily/string.h>
 #define	__XDEV__	/* Needed to activate _dev_major()/_dev_minor() */
 #include <schily/device.h>
+#define	GT_COMERR		/* #define comerr gtcomerr */
+#define	GT_ERROR		/* #define error gterror   */
 #include <schily/schily.h>
 #include <schily/idcache.h>
 #include "starsubs.h"
 #include "movearch.h"
 #include "xtab.h"
+#include "pathname.h"
 
 extern	BOOL	no_xheader;
 extern	BOOL	nowarn;
+extern	BOOL	binflag;
 
 extern	GINFO	*grip;				/* Global read info pointer	*/
 
@@ -106,6 +112,7 @@ LOCAL	void	get_realsize	__PR((FINFO *info, char *keyword, int klen, char *arg, i
 LOCAL	void	get_offset	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_major	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_minor	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
+LOCAL	void	get_minorbits	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_dev		__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_ino		__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_nlink	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
@@ -114,6 +121,7 @@ LOCAL	void	get_filetype	__PR((FINFO *info, char *keyword, int klen, char *arg, i
 LOCAL	void	get_acl_type	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_acl_access	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_acl_default	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
+LOCAL	void	get_acl_ace	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 #endif
 #ifdef  USE_XATTR
 LOCAL	void	get_attr	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
@@ -125,7 +133,9 @@ LOCAL	void	get_dir		__PR((FINFO *info, char *keyword, int klen, char *arg, int l
 LOCAL	void	get_iarray	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_release	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_archtype	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
+LOCAL	void	get_hdrcharset	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
 LOCAL	void	get_dummy	__PR((FINFO *info, char *keyword, int klen, char *arg, int len));
+LOCAL	void	unsup_arg	__PR((char *keyword, char *arg));
 LOCAL	void	bad_utf8	__PR((char *keyword, char *arg));
 
 /*
@@ -142,7 +152,7 @@ LOCAL	Uchar	dtab[] = "0123456789";
 					(val) = (val) / (unsigned)10;	      \
 				} while ((val) > 0)
 
-#define	scopy(to, from)		while ((*(to)++ = *(from)++) != '\0');
+#define	scopy(to, from)		while ((*(to)++ = *(from)++) != '\0')
 
 
 LOCAL	char	*xbuf;	/* Space used to prepare I/O from/to extended headers */
@@ -180,7 +190,7 @@ LOCAL xtab_t xtab[] = {
 
 			{ "charset",		7, get_dummy,	0	},
 			{ "comment",		7, get_dummy,	0	},
-			{ "hdrcharset",		10, get_dummy,	0	},
+			{ "hdrcharset",		10, get_hdrcharset,	0	},
 /* "BINARY" */
 
 			{ "SCHILY.devmajor",	15, get_major,	0	},
@@ -189,6 +199,7 @@ LOCAL xtab_t xtab[] = {
 #ifdef	USE_ACL
 			{ "SCHILY.acl.access",	17, get_acl_access, 0	},
 			{ "SCHILY.acl.default",	18, get_acl_default, 0	},
+			{ "SCHILY.acl.ace",	14, get_acl_ace, 0	},
 			{ "SCHILY.acl.type",	15, get_acl_type, 0	},
 #else
 /*
@@ -197,6 +208,7 @@ LOCAL xtab_t xtab[] = {
  */
 			{ "SCHILY.acl.access",	17, get_dummy,	0	},
 			{ "SCHILY.acl.default",	18, get_dummy,	0	},
+			{ "SCHILY.acl.ace",	14, get_dummy,	0	},
 			{ "SCHILY.acl.type",	15, get_dummy,	0	},
 #endif
 #ifdef  USE_XATTR
@@ -215,6 +227,7 @@ LOCAL xtab_t xtab[] = {
 			{ "SCHILY.fflags",	13, get_dummy,	0	},
 #endif
 			{ "SCHILY.dev",		10, get_dev,	0	},
+			{ "SCHILY.devminorbits", 19, get_minorbits,	0	},
 			{ "SCHILY.ino",		10, get_ino,	0	},
 			{ "SCHILY.nlink",	12, get_nlink,	0	},
 			{ "SCHILY.filetype",	15, get_filetype, 0	},
@@ -251,14 +264,19 @@ xbinit()
 {
 	_xbinit();
 
-	ginfo.f_name = ___malloc(PATH_MAX+1, "global info name");
-	ginfo.f_lname = ___malloc(PATH_MAX+1, "global info lname");
+	init_pspace(PS_EXIT, &ginfo.f_pname);
+	ginfo.f_name = ginfo.f_pname.ps_path;
 	ginfo.f_name[0] = '\0';
+
+	init_pspace(PS_EXIT, &ginfo.f_plname);
+	ginfo.f_lname = ginfo.f_plname.ps_path;
 	ginfo.f_lname[0] = '\0';
+
 	ginfo.f_uname = ___malloc(MAX_UNAME+1, "global user name");
 	ginfo.f_gname = ___malloc(MAX_UNAME+1, "global group name");
 	ginfo.f_uname[0] = '\0';
 	ginfo.f_gname[0] = '\0';
+	ginfo.f_devminorbits = 0;
 	ggname = ginfo.f_gname;
 	ginfo.f_xflags = 0;
 	guname = ginfo.f_uname;
@@ -388,6 +406,8 @@ extern	BOOL	dodump;
 	 */
 	if ((xflags & (XF_ATIME|XF_CTIME|XF_MTIME|XF_NOTIME)) == 0)
 		xflags |= (XF_ATIME|XF_CTIME|XF_MTIME);
+	else if (xflags & XF_NOTIME)
+		xflags &= ~(XF_ATIME|XF_CTIME|XF_MTIME);
 
 #ifdef	DEBUG_XHDR
 	xflags = 0xffffffff;
@@ -494,7 +514,14 @@ extern	BOOL	dodump;
 	if (xflags & (XF_ACL_ACCESS|XF_ACL_DEFAULT)) {
 		gen_text("SCHILY.acl.type", "POSIX draft", 11, 0);
 	}
+	if (xflags & XF_ACL_ACE) {
+		gen_text("SCHILY.acl.type", "NFSv4", 5, 0);
+	}
 #endif
+	if (xflags & XF_ACL_ACE) {
+		gen_text("SCHILY.acl.ace", info->f_acl_ace, -1, T_UTF8);
+	}
+
 	if (xflags & XF_ACL_ACCESS) {
 		gen_text("SCHILY.acl.access", info->f_acl_access, -1, T_UTF8);
 	}
@@ -966,7 +993,7 @@ gen_text(keyword, arg, alen, flags)
 
 	i = len;
 	if (flags & T_UTF8)
-		i *= 2;			/* UTF-8 Factor may be up to 6!	    */
+		i *= 6;			/* UTF-8 Factor may be up to 6!	    */
 	if ((xbidx + i) > xblen)
 		xbgrow(i);
 
@@ -977,8 +1004,8 @@ gen_text(keyword, arg, alen, flags)
 	--p,			/* Overshoot compensation		    */
 	*p++ = '=';		/* Fill in '=' after keyword field	    */
 
-	if (flags & T_UTF8) {
-		i = to_utf8l((Uchar *)p, (Uchar *)arg, alen);
+	if (flags & T_UTF8 && !binflag) {
+		i = to_utf8((Uchar *)p, i, (Uchar *)arg, alen);
 		p += i + 1;
 	} else {
 		p = movebytes(arg, p, alen);
@@ -1377,7 +1404,7 @@ get_xtime(keyword, arg, len, secp, nsecp)
 	long	*nsecp;
 {
 #ifdef	__use_default_time__
-extern struct	timeval	ddate;
+extern struct	timespec	ddate;
 #endif
 	Llong	ll;
 	long	l;
@@ -1670,8 +1697,8 @@ get_gid(info, keyword, klen, arg, len)
  * Space for returning user/group names.
  * XXX If we ever change to use allocated space, we need to change info_xcopy()
  */
-LOCAL	Uchar	_uname[MAX_UNAME+1];
-LOCAL	Uchar	_gname[MAX_UNAME+1];
+LOCAL	Uchar	_uname[MAX_UNAME+2];
+LOCAL	Uchar	_gname[MAX_UNAME+2];
 
 /*
  * get user name (if name length is > 32 chars or if contains non ASCII chars)
@@ -1689,14 +1716,19 @@ get_uname(info, keyword, klen, arg, len)
 		info->f_xflags &= ~XF_UNAME;
 		return;
 	}
-	if (strlen(arg) > MAX_UNAME) {
+	if (len > MAX_UNAME) {
 		print_toolong(keyword, arg, len);
 		return;
 	}
-	if (from_utf8(_uname, (Uchar *)arg)) {
+	if (ginfo.f_xflags & XF_BINARY) {
+		strcpy((char *)_uname, arg);
 		info->f_xflags |= XF_UNAME;
 		info->f_uname = (char *)_uname;
-		info->f_umaxlen = strlen((char *)_uname);
+		info->f_umaxlen = len;
+	} else if (from_utf8(_uname, sizeof (_uname), (Uchar *)arg, &len)) {
+		info->f_xflags |= XF_UNAME;
+		info->f_uname = (char *)_uname;
+		info->f_umaxlen = len;
 	} else {
 		bad_utf8(keyword, arg);
 	}
@@ -1718,14 +1750,19 @@ get_gname(info, keyword, klen, arg, len)
 		info->f_xflags &= ~XF_GNAME;
 		return;
 	}
-	if (strlen(arg) > MAX_UNAME) {
+	if (len > MAX_UNAME) {
 		print_toolong(keyword, arg, len);
 		return;
 	}
-	if (from_utf8(_gname, (Uchar *)arg)) {
+	if (ginfo.f_xflags & XF_BINARY) {
+		strcpy((char *)_gname, arg);
 		info->f_xflags |= XF_GNAME;
 		info->f_gname = (char *)_gname;
-		info->f_gmaxlen = strlen((char *)_gname);
+		info->f_gmaxlen = len;
+	} else if (from_utf8(_gname, sizeof (_gname), (Uchar *)arg, &len)) {
+		info->f_xflags |= XF_GNAME;
+		info->f_gname = (char *)_gname;
+		info->f_gmaxlen = len;
 	} else {
 		bad_utf8(keyword, arg);
 	}
@@ -1747,19 +1784,53 @@ get_path(info, keyword, klen, arg, len)
 		info->f_xflags &= ~XF_PATH;
 		return;
 	}
-	if (strlen(arg) > PATH_MAX) {
-		print_toolong(keyword, arg, len);
-		return;
-	}
+
 	/*
 	 * Check whether we are called via get_xhtype() -> xhparse()
 	 */
 	if (info->f_name == NULL)
 		return;
-	if (from_utf8((Uchar *)info->f_name, (Uchar *)arg)) {
+
+	if (ginfo.f_xflags & XF_BINARY) {
+		if (strlcpy_pspace(PS_STDERR, &info->f_pname, arg, len) < 0) {
+			print_toolong(keyword, arg, len);
+			info->f_xflags |= F_BAD_META;
+			return;
+		}
+		info->f_name = info->f_pname.ps_path;
+		info->f_namelen = len;
 		info->f_xflags |= XF_PATH;
 	} else {
-		bad_utf8(keyword, arg);
+		int	ilen = len;
+		BOOL	ret;
+
+		do {
+			len = ilen;
+			ret = from_utf8((Uchar *)info->f_name,
+					info->f_pname.ps_size,
+					(Uchar *)arg, &len);
+			if (len >= info->f_pname.ps_size) {
+				/*
+				 * An increment of 1 is OK, since it is unlikely
+				 * that the path grows by more than 256 per dir.
+				 */
+				if (incr_pspace(PS_STDERR,
+						&info->f_pname, 1) < 0) {
+					print_toolong(keyword, arg, len);
+					info->f_xflags |= F_BAD_META;
+					return;
+				}
+				info->f_name = info->f_pname.ps_path;
+			} else
+				break;
+		} while (1);
+
+		if (ret) {
+			info->f_namelen = len;
+			info->f_xflags |= XF_PATH;
+		} else {
+			bad_utf8(keyword, arg);
+		}
 	}
 }
 
@@ -1779,19 +1850,53 @@ get_lpath(info, keyword, klen, arg, len)
 		info->f_xflags &= ~XF_LINKPATH;
 		return;
 	}
-	if (strlen(arg) > PATH_MAX) {
-		print_toolong(keyword, arg, len);
-		return;
-	}
+
 	/*
 	 * Check whether we are called via get_xhtype() -> xhparse()
 	 */
 	if (info->f_lname == NULL)
 		return;
-	if (from_utf8((Uchar *)info->f_lname, (Uchar *)arg)) {
+
+	if (ginfo.f_xflags & XF_BINARY) {
+		if (strlcpy_pspace(PS_STDERR, &info->f_plname, arg, len) < 0) {
+			print_toolong(keyword, arg, len);
+			info->f_xflags |= F_BAD_META;
+			return;
+		}
+		info->f_lname = info->f_plname.ps_path;
+		info->f_lnamelen = len;
 		info->f_xflags |= XF_LINKPATH;
 	} else {
-		bad_utf8(keyword, arg);
+		int	ilen = len;
+		BOOL	ret;
+
+		do {
+			len = ilen;
+			ret = from_utf8((Uchar *)info->f_lname,
+					info->f_plname.ps_size,
+					(Uchar *)arg, &len);
+			if (len >= info->f_plname.ps_size) {
+				/*
+				 * An increment of 1 is OK, since it is unlikely
+				 * that the path grows by more than 256 per dir.
+				 */
+				if (incr_pspace(PS_STDERR,
+						&info->f_plname, 1) < 0) {
+					print_toolong(keyword, arg, len);
+					info->f_xflags |= F_BAD_META;
+					return;
+				}
+				info->f_lname = info->f_plname.ps_path;
+			} else
+				break;
+		} while (1);
+
+		if (ret) {
+			info->f_lnamelen = len;
+			info->f_xflags |= XF_LINKPATH;
+		} else {
+			bad_utf8(keyword, arg);
+		}
 	}
 }
 
@@ -1970,6 +2075,34 @@ get_minor(info, keyword, klen, arg, len)
 	}
 }
 
+/*
+ * get number of minor bits for this file (vendor unique)
+ * This is naturally unsigned.
+ */
+/* ARGSUSED */
+LOCAL void
+get_minorbits(info, keyword, klen, arg, len)
+	FINFO	*info;
+	char	*keyword;
+	int	klen;
+	char	*arg;
+	int	len;
+{
+	Ullong	ull;
+
+	if (len == 0) {
+		info->f_devminorbits = 0;
+		return;
+	}
+	if (get_unumber(keyword, arg, &ull, INO_T_MAX)) {
+		info->f_devminorbits = ull;
+		if (info->f_devminorbits != ull) {
+			xh_rangeerr(keyword, arg, len);
+			info->f_devminorbits = 0;
+		}
+	}
+}
+
 /*
  * get device number of device containing FS (vendor unique)
  * The device number should be unsigned but POSIX does not say anything
@@ -1987,16 +2120,56 @@ get_dev(info, keyword, klen, arg, len)
 	Ullong	ull;
 	BOOL	neg = FALSE;
 
+	info->f_devmaj = 0;
+	info->f_devmin = 0;
 	if (len == 0) {
 		info->f_dev = 0;
 		return;
 	}
 	if (get_snumber(keyword, arg, &ull, &neg,
 					-(Ullong)DEV_T_MIN, DEV_T_MAX)) {
+		int	mbits;
+		Llong	ll;
+
 		if (neg)
-			info->f_dev = -ull;
+			info->f_dev = ll = -ull;
 		else
-			info->f_dev = ull;
+			info->f_dev = ll = ull;
+
+		mbits = info->f_devminorbits;
+		if (mbits == 0)
+			mbits = ginfo.f_devminorbits;
+		if (mbits) {
+			int	oerr = geterrno();
+
+			seterrno(0);
+			info->f_devmaj	= _dev_major(mbits, ll);
+			info->f_devmin	= _dev_minor(mbits, ll);
+			info->f_dev = makedev(info->f_devmaj, info->f_devmin);
+			ull = dev_make(info->f_devmaj, info->f_devmin);
+			/*
+			 * Check whether the device from the archive
+			 * can be represented on the local system.
+			 */
+			if (geterrno() ||
+			    info->f_devmaj != major(info->f_dev) ||
+			    info->f_devmin != minor(info->f_dev) ||
+			    (Ullong)info->f_dev != ull) {
+				xh_rangeerr(keyword, arg, len);
+				info->f_dev = 0;
+				info->f_devmaj = 0;
+				info->f_devmin = 0;
+			}
+			seterrno(oerr);
+			return;
+		}
+		/*
+		 * Let us hope that both, the archiving and the
+		 * extracting system use the same major()/minor()
+		 * mapping.
+		 */
+		info->f_devmaj	= major(ll);
+		info->f_devmin	= minor(ll);
 
 		if ((neg && -info->f_dev != ull) ||
 			(!neg && info->f_dev != ull)) {
@@ -2085,11 +2258,13 @@ get_filetype(info, keyword, klen, arg, len)
 		return;
 	}
 
-	for (i = 0; i < XT_BAD; i++) {
+	for (i = 0; i <= XT_BAD; i++) {
+		if (xtnamelen_tab[i] != len)
+			continue;
 		if (streql(xttoname_tab[i], arg))
 			break;
 	}
-	if (i >= XT_BAD)
+	if (i >= XT_BAD)			/* Use type from tarhdr */
 		return;
 
 	if (keyword[7] == 'f') {		/* "SCHILY.filetype" */
@@ -2106,8 +2281,8 @@ get_filetype(info, keyword, klen, arg, len)
 /*
  * XXX acl_access_text/acl_default_text are a bad idea. (see acl_unix.c)
  */
-LOCAL char acl_access_text[PATH_MAX+1];
-LOCAL char acl_default_text[PATH_MAX+1];
+LOCAL pathstore_t	acl_access_text;
+LOCAL pathstore_t	acl_default_text;
 
 /* ARGSUSED */
 LOCAL void
@@ -2120,6 +2295,8 @@ get_acl_type(info, keyword, klen, arg, len)
 {
 	if (len == 11 && streql(arg, "POSIX draft"))
 		return;
+	if (len == 5 && streql(arg, "NFSv4"))
+		return;
 
 	info->f_flags |= F_BAD_ACL;
 
@@ -2145,11 +2322,19 @@ get_acl_access(info, keyword, klen, arg, len)
 		info->f_acl_access = NULL;
 		return;
 	}
-	if (strlen(arg) > PATH_MAX)	/* XXX We should use dynamic strings */
+	if ((len + 2) > acl_access_text.ps_size)
+		grow_pspace(PS_EXIT, &acl_access_text, (len + 2));
+	if (acl_access_text.ps_path == NULL)
 		return;
-	if (from_utf8((Uchar *)acl_access_text, (Uchar *)arg)) {
+	if (ginfo.f_xflags & XF_BINARY) {
+		strcpy(acl_access_text.ps_path, arg);
 		info->f_xflags |= XF_ACL_ACCESS;
-		info->f_acl_access = acl_access_text;
+		info->f_acl_access = acl_access_text.ps_path;
+	} else if (from_utf8((Uchar *)acl_access_text.ps_path,
+		    acl_access_text.ps_size,
+		    (Uchar *)arg, &len)) {
+		info->f_xflags |= XF_ACL_ACCESS;
+		info->f_acl_access = acl_access_text.ps_path;
 	} else {
 		bad_utf8(keyword, arg);
 	}
@@ -2169,11 +2354,53 @@ get_acl_default(info, keyword, klen, arg, len)
 		info->f_acl_default = NULL;
 		return;
 	}
-	if (strlen(arg) > PATH_MAX)	/* XXX We should use dynamic strings */
+	if ((len + 2) > acl_default_text.ps_size)
+		grow_pspace(PS_EXIT, &acl_default_text, (len + 2));
+	if (acl_default_text.ps_path == NULL)
 		return;
-	if (from_utf8((Uchar *)acl_default_text, (Uchar *)arg)) {
+	if (ginfo.f_xflags & XF_BINARY) {
+		strcpy(acl_default_text.ps_path, arg);
 		info->f_xflags |= XF_ACL_DEFAULT;
-		info->f_acl_default = acl_default_text;
+		info->f_acl_default = acl_default_text.ps_path;
+	} else if (from_utf8((Uchar *)acl_default_text.ps_path,
+		    acl_default_text.ps_size,
+		    (Uchar *)arg, &len)) {
+		info->f_xflags |= XF_ACL_DEFAULT;
+		info->f_acl_default = acl_default_text.ps_path;
+	} else {
+		bad_utf8(keyword, arg);
+	}
+}
+
+LOCAL pathstore_t	acl_ace_text;
+
+/* ARGSUSED */
+LOCAL void
+get_acl_ace(info, keyword, klen, arg, len)
+	FINFO	*info;
+	char	*keyword;
+	int	klen;
+	char	*arg;
+	int	len;
+{
+	if (len == 0 || (info->f_flags & F_BAD_ACL)) {
+		info->f_xflags &= ~XF_ACL_ACE;
+		info->f_acl_ace = NULL;
+		return;
+	}
+	if ((len + 2) > acl_ace_text.ps_size)
+		grow_pspace(PS_EXIT, &acl_ace_text, (len + 2));
+	if (acl_ace_text.ps_path == NULL)
+		return;
+	if (ginfo.f_xflags & XF_BINARY) {
+		strcpy(acl_ace_text.ps_path, arg);
+		info->f_xflags |= XF_ACL_ACE;
+		info->f_acl_ace = acl_ace_text.ps_path;
+	} else if (from_utf8((Uchar *)acl_ace_text.ps_path,
+		    acl_ace_text.ps_size,
+		    (Uchar *)arg, &len)) {
+		info->f_xflags |= XF_ACL_ACE;
+		info->f_acl_ace = acl_ace_text.ps_path;
 	} else {
 		bad_utf8(keyword, arg);
 	}
@@ -2283,10 +2510,14 @@ get_dir(info, keyword, klen, arg, len)
 	if (arg[len-2] != '\0')
 		arg[len++] = '\0';	/* Kill '\n' */
 
-	/*
-	 * The non UTF-8 string is shorter so we convert in place.
-	 */
-	from_utf8l((Uchar *)arg, (Uchar *)arg, &len);
+	if (ginfo.f_xflags & XF_BINARY) {
+		;
+	} else {
+		/*
+		 * The non UTF-8 string is shorter so we convert in place.
+		 */
+		from_utf8((Uchar *)arg, len, (Uchar *)arg, &len);
+	}
 	info->f_dir = arg;
 	info->f_dirlen = len;
 }
@@ -2373,10 +2604,14 @@ get_release(info, keyword, klen, arg, len)
 		grip->gflags &= ~GF_RELEASE;
 		return;
 	}
-	/*
-	 * The non UTF-8 string is shorter so we convert in place.
-	 */
-	from_utf8l((Uchar *)arg, (Uchar *)arg, &len);
+	if (ginfo.f_xflags & XF_BINARY) {
+		;
+	} else {
+		/*
+		 * The non UTF-8 string is shorter so we convert in place.
+		 */
+		from_utf8((Uchar *)arg, len, (Uchar *)arg, &len);
+	}
 	grip->gflags |= GF_RELEASE;
 	grip->release = ___savestr(arg);
 }
@@ -2399,6 +2634,31 @@ get_archtype(info, keyword, klen, arg, len)
 	grip->archtype = hdr_type(arg);
 }
 
+/*
+ * Get the charset used for path, linkpath, uname and gname.
+ */
+/* ARGSUSED */
+LOCAL void
+get_hdrcharset(info, keyword, klen, arg, len)
+	FINFO	*info;
+	char	*keyword;
+	int	klen;
+	char	*arg;
+	int	len;
+{
+#ifdef	__needed__
+	BOOL	is_global = info == &ginfo;
+#endif
+
+	if (len == 23 && streql("ISO-IR 10646 2000 UTF-8", arg)) {
+		info->f_xflags &= ~XF_BINARY;
+	} else if (len == 6 && streql("BINARY", arg)) {
+		info->f_xflags |= XF_BINARY;
+	} else {
+		unsup_arg(keyword, arg);
+	}
+}
+
 /*
  * Dummy 'get' function used for all fields that we don't yet understand or
  * fields that we indent to ignore.
@@ -2414,6 +2674,16 @@ get_dummy(info, keyword, klen, arg, len)
 {
 }
 
+LOCAL void
+unsup_arg(keyword, arg)
+	char	*keyword;
+	char	*arg;
+{
+	errmsgno(EX_BAD,
+		"Unsupported arg '%s' for '%s' in extended header at %lld.\n",
+		arg, keyword, tblocks());
+}
+
 LOCAL void
 bad_utf8(keyword, arg)
 	char	*keyword;
